# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from django.core.exceptions import ValidationError
from django.db.utils import IntegrityError
from django.utils import timezone

from swh.web.alter.models import (
    Alteration,
    AlterationCategory,
    AlterationStatus,
    BlockList,
    Event,
    EventCategory,
    Origin,
    Token,
    validate_email_or_domain,
)


def test_origin_str(alteration):
    origin = alteration.origins.first()
    assert str(origin) == origin.url


def test_origin_admin_url(alteration):
    origin = alteration.origins.first()
    assert (
        origin.get_admin_url()
        == f"/admin/alteration/{alteration.pk}/origin/{origin.pk}/"
    )


def test_unique_origin(alteration):
    origin = alteration.origins.first()
    with pytest.raises(IntegrityError, match="unique_url"):
        Origin.objects.create(url=origin.url, alteration=alteration)


def test_alteration_search(db):
    requesters = [
        ("dona@gmail.com", "http://github.com/centipede"),
        ("bug@mark.two", "http://gitlab.com/cobol"),
        ("switched@on.bach", "http://gitlab.com/moog"),
    ]
    for email, url in requesters:
        alteration = Alteration.objects.create(
            reasons="Reasons",
            expected_outcome="takedown everything",
            email=email,
        )
        Origin.objects.create(url=url, alteration=alteration)

    bailey = Alteration.objects.search("gmail")  # matches by email
    assert bailey.count() == 1
    assert bailey.first().email == "dona@gmail.com"

    gitlab = Alteration.objects.search("GitLab")  # matches by origin url
    assert gitlab.count() == 2

    outcome = Alteration.objects.search("takedown")  # matches by outcome
    assert outcome.count() == 3

    nothing = Alteration.objects.search("ZZZ")  # matches nothing
    assert nothing.count() == 0


def test_alteration_admin_url(alteration):
    assert alteration.get_admin_url() == f"/admin/alteration/{alteration.pk}/"


def test_alteration_absolute_url(alteration):
    assert alteration.get_absolute_url() == f"/alteration/{alteration.pk}/"


def test_alteration_create_from_assistant(client, db):
    session = client.session
    session["alteration_category"] = AlterationCategory.COPYRIGHT
    session["alteration_reasons"] = "My reasons"
    session["alteration_expected_outcome"] = "Expected outcome"
    session["alteration_email"] = "my@email.address"
    session["alteration_origins"] = ["https://xerox.com/vlsi"]

    alteration = Alteration.create_from_assistant(session)

    assert alteration.category == AlterationCategory.COPYRIGHT
    assert alteration.reasons == "My reasons"
    assert alteration.expected_outcome == "Expected outcome"
    assert alteration.email == "my@email.address"
    origins = [o.url for o in alteration.origins.all()]
    assert origins == ["https://xerox.com/vlsi"]

    assert alteration.events.count() == 1
    event = alteration.events.first()

    assert event.category == EventCategory.LOG
    assert event.author == "Requester"
    assert "created" in event.content
    assert not event.internal


def test_alteration_ro(alteration):
    for status in AlterationStatus:
        alteration.status = status
        if status == AlterationStatus.ARCHIVED:
            assert alteration.is_read_only
        else:
            assert not alteration.is_read_only


def test_alteration_str(alteration):
    assert alteration.get_category_display() in str(alteration)
    assert alteration.email in str(alteration)


def test_event_public_manager(alteration, message_event, admin_message_event):
    assert alteration.events.count() == 4  # 2 logs + 2 messages
    assert Event.public_objects.filter(alteration=alteration).count() == 2
    assert Event.public_objects.filter(alteration=alteration).first() == message_event


def test_event_str(admin_message_event):
    assert admin_message_event.get_category_display() in str(admin_message_event)
    assert admin_message_event.author in str(admin_message_event)
    assert str(admin_message_event.alteration) in str(admin_message_event)


def test_alteration_token_absolute_url(token):
    assert token.get_absolute_url() == f"/alteration/link/{token.value}/"


def test_token_unique(token):
    alteration = Alteration.objects.create(
        reasons="Other reasons",
        expected_outcome="takedown",
        email="jm@email.address",
    )
    with pytest.raises(IntegrityError, match="unique_token_value"):
        Token.objects.create(value=token.value, alteration=alteration)


def test_token_create_for_alteration(alteration):
    token = Token.create_for(alteration)
    assert token.alteration == alteration
    assert token.email is None


@pytest.mark.django_db
def test_token_create_for_email():
    token = Token.create_for("test@email.local")
    assert token.email == "test@email.local"
    assert token.alteration is None


@pytest.mark.django_db
def test_token_create_for_something_else():
    with pytest.raises(ValueError, match="Invalid parameter"):
        Token.create_for(123)


@pytest.mark.django_db
def test_email_token_absolute_url():
    token = Token.create_for("test@email.local")
    assert token.get_absolute_url() == f"/alteration/email/verification/{token.value}/"


def test_token_expired(token):
    assert not token.expired
    token.expires_at = timezone.now()
    assert token.expired


@pytest.mark.skip(
    reason="FIXME: can't find a way to patch Token.value's default method"
)
def test_token_create_for_exc(alteration, token, mocker):
    # side_effect = [token.value] * 5
    # mocker.patch(
    #     "swh.web.alter.models._default_token_value",
    #     side_effect=side_effect,
    # )
    # Token.value.field.default = mocker.Mock(side_effect=side_effect)
    # mocker.patch(
    #     "swh.web.alter.models.Token.value.field.default",
    #     side_effect=side_effect,
    # )
    with pytest.raises(IntegrityError, match="after 5 attempts"):
        Token.create_for_alteration(alteration)


@pytest.mark.django_db
@pytest.mark.parametrize(
    "value,expected_failure",
    [
        ("email@domain.localhost", False),
        ("domain.localhost", False),
        ("email@", True),
        ("localhost.%", True),
    ],
)
def test_validate_email_or_domain(value, expected_failure):
    if expected_failure:
        with pytest.raises(ValidationError):
            validate_email_or_domain(value)
    else:
        try:
            validate_email_or_domain(value)
        except ValidationError:
            pytest.fail("Unexpected error")


@pytest.mark.django_db
def test_block_list_unique():
    BlockList.objects.create(email_or_domain="test.localhost")
    with pytest.raises(IntegrityError, match="unique_email_or_domain"):
        BlockList.objects.create(email_or_domain="test.localhost")


@pytest.mark.django_db
def test_block_list_is_blocked():
    BlockList.objects.create(email_or_domain="test.localhost")
    with pytest.raises(ValueError, match="Invalid"):
        BlockList.is_blocked("test.localhost")  # expects a valid email address
    assert BlockList.is_blocked("whatever@test.localhost")
    assert not BlockList.is_blocked("whatever@test.other")


@pytest.mark.django_db
def test_block_list_lowercased():
    block = BlockList.objects.create(email_or_domain="TEST.localhost")
    assert block.email_or_domain == "test.localhost"


@pytest.mark.django_db
def test_block_list_disposable(settings):
    from disposable_email_domains import blocklist

    disposable_domain = next(iter(blocklist))
    email = f"test@{disposable_domain}"
    settings.ALTER_SETTINGS["block_disposable_email_domains"] = False
    assert not BlockList.is_blocked(email)

    settings.ALTER_SETTINGS["block_disposable_email_domains"] = True
    assert BlockList.is_blocked(email)



# Copyright (C) 2022-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from dataclasses import dataclass
from io import BytesIO, StringIO
import sys

import pytest

from django.core.management import call_command

from .common_tests import *  # noqa


class MockedStdin:
    def __init__(self):
        self.buffer = BytesIO()


@dataclass
class CommandReturn:
    out: str
    err: str


@pytest.fixture
def receive_inbound_message(inbound_message):
    def do_it(expect_failure=False):
        orig_stdin = sys.stdin
        try:
            sys.stdin = MockedStdin()
            sys.stdin.buffer.write(inbound_message)
            sys.stdin.buffer.seek(0)

            out = StringIO()
            err = StringIO()

            call_command("process_inbound_email", stdout=out, stderr=err)

            out.seek(0)
            err.seek(0)

            assert out.read() == ""
            assert err.read() == ""
        finally:
            sys.stdin = orig_stdin

        return inbound_message

    return do_it


@pytest.mark.inbound_message(b"")
def test_empty_stdin(caplog, receive_inbound_message):
    receive_inbound_message()

    assert len(caplog.records) == 1
    log = caplog.records[0]
    assert log.levelname == "ERROR"
    assert "Unhandled message" in log.getMessage()



# Copyright (C) 2022-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import logging
from typing import Any, Dict

from django_js_reverse.views import urls_js
import requests

from django.shortcuts import render
from django.templatetags.static import static
from django.urls import re_path as url
from django.views.generic.base import RedirectView

from swh.web.config import get_config
from swh.web.utils import archive, django_cache, origin_visit_types
from swh.web.utils.exc import sentry_capture_exception

swh_web_config = get_config()

SWH_FAVICON = "img/icons/swh-logo-32x32.png"

favicon_view = RedirectView.as_view(url=static(SWH_FAVICON), permanent=True)

logger = logging.getLogger(__name__)


@django_cache()
def _stat_counters() -> Dict[str, Any]:
    stat_counters = {}
    stat_counters_history = {}
    try:
        stat_counters = archive.stat_counters()
        url = get_config()["history_counters_url"]
        if url:
            response = requests.get(url, timeout=5)
            stat_counters_history = json.loads(response.text)

    except Exception as exc:
        sentry_capture_exception(exc)

    return {
        "stat_counters": stat_counters,
        "stat_counters_history": stat_counters_history,
    }


def default_view(request):
    return render(
        request,
        "homepage.html",
        {"visit_types": origin_visit_types(use_cache=True), **_stat_counters()},
    )


urlpatterns = [
    url(r"^favicon\.ico/$", favicon_view, name="favicon"),
    # to prevent django.template.base.VariableDoesNotExist exception when
    # browsing django admin site
    url(r"^favicon\.ico$", favicon_view, name="favicon-no-trailing-slash"),
    url(r"^$", default_view, name="swh-web-homepage"),
    url(r"^jsreverse/$", urls_js, name="js-reverse"),
]



# Copyright (C) 2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.web.tests.helpers import check_api_post_response
from swh.web.utils import reverse


def test_add_origin_with_contents(api_client, archive_data):
    origin_url = "https://git.example.org/simple_repo"
    contents = [{"path": "test/bar", "data": "foo"}]
    url = reverse(
        "tests-add-origin-with-contents", query_params={"origin_url": origin_url}
    )
    check_api_post_response(api_client, url, data=contents, status_code=200)
    assert archive_data.origin_get([origin_url])
    snapshot = archive_data.snapshot_get_latest(origin_url)
    revision = archive_data.revision_get(snapshot["branches"]["main"]["target"])
    entries = archive_data.directory_ls(revision["directory"], recursive=True)
    assert {entry["name"] for entry in entries} == {"test", "test/bar"}



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json

from django.contrib.staticfiles import finders
from django.shortcuts import render
from django.urls import re_path as url


def jslicenses(request):
    jslicenses_file = finders.find("jssources/jslicenses.json")
    jslicenses_data = {}
    if jslicenses_file is not None:
        jslicenses_data = json.load(open(jslicenses_file))
    jslicenses_data = sorted(
        jslicenses_data.items(), key=lambda item: item[0].split("/")[-1]
    )
    return render(request, "jslicenses.html", {"jslicenses_data": jslicenses_data})


urlpatterns = [
    url(r"^jslicenses/$", jslicenses, name="jslicenses"),
]



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import unicode_literals

from django.db import migrations, models

from swh.web.config import scheduler


def _remove_archived_tasks_with_no_saved_status(apps, schema_editor):
    """
    Scheduler tasks are archived on a regular basis so their completion
    state could not be known anymore as previous to this migration,
    the loading task status was not stored in the database.
    So remove the rows associated to already archived tasks as
    the loading status cannot be retrieved anymore.
    """
    SaveOriginRequest = apps.get_model("swh_web_save_code_now", "SaveOriginRequest")
    no_saved_status_tasks = []
    for sor in SaveOriginRequest.objects.all():
        tasks = scheduler().get_tasks([sor.loading_task_id])
        if not tasks:
            no_saved_status_tasks.append(sor.loading_task_id)
    SaveOriginRequest.objects.filter(loading_task_id__in=no_saved_status_tasks).delete()


class Migration(migrations.Migration):
    dependencies = [
        ("swh_web_save_code_now", "0002_saveoriginrequest_visit_date"),
    ]

    operations = [
        migrations.AddField(
            model_name="saveoriginrequest",
            name="loading_task_status",
            field=models.TextField(
                choices=[
                    ("not created", "not created"),
                    ("not yet scheduled", "not yet scheduled"),
                    ("scheduled", "scheduled"),
                    ("succeed", "succeed"),
                    ("failed", "failed"),
                ],
                default="not created",
            ),
        ),
        migrations.RunPython(_remove_archived_tasks_with_no_saved_status),
    ]



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import sys

from django.core.management.base import BaseCommand

from ...handle_message import MessageHandler


class Command(BaseCommand):
    help = "Process a new inbound email"

    def handle(self, *args, **options):
        MessageHandler(
            raw_message=sys.stdin.buffer.read(), sender=self.__class__
        ).handle()



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("swh_web_auth", "0006_fix_mailmap_admin_user_id"),
    ]

    operations = [
        # as we simply move the mailmap feature to a dedicated django application,
        # we do not want to remove the tables in database to not lose data
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.DeleteModel(
                    name="UserMailmap",
                ),
                migrations.DeleteModel(
                    name="UserMailmapEvent",
                ),
            ],
            database_operations=[],
        ),
    ]



# Copyright (C) 2022-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime, timedelta, timezone
from typing import Any, Dict

import iso8601

from swh.web.save_code_now.models import SaveOriginRequest
from swh.web.save_code_now.origin_save import (
    SAVE_TASK_SUCCEEDED,
    WEBHOOK_REQUEST_COOLDOWN_INTERVAL,
    get_save_origin_request,
)
from swh.web.tests.helpers import check_api_post_response
from swh.web.utils import reverse


def django_http_headers(http_headers: Dict[str, Any]):
    return {f"HTTP_{k.upper().replace('-', '_')}": v for k, v in http_headers.items()}


def origin_save_webhook_receiver_test(
    forge_type: str,
    http_headers: Dict[str, Any],
    payload: Dict[str, Any],
    expected_origin_url: str,
    expected_visit_type: str,
    api_client,
    swh_scheduler,
):
    url = reverse(f"api-1-origin-save-webhook-{forge_type.lower()}")

    resp = check_api_post_response(
        api_client,
        url,
        status_code=200,
        data=payload,
        **django_http_headers(http_headers),
    )

    assert resp.data["origin_url"] == expected_origin_url
    assert resp.data["visit_type"] == expected_visit_type
    assert "id" in resp.data
    assert resp.data["request_url"] == reverse(
        "api-1-save-origin",
        url_args={"request_id": resp.data["id"]},
        request=resp.wsgi_request,  # type: ignore
    )

    tasks = swh_scheduler.search_tasks(task_type=f"load-{expected_visit_type}")
    assert tasks
    assert tasks[0].arguments.kwargs["url"] == expected_origin_url

    request = SaveOriginRequest.objects.get(
        origin_url=expected_origin_url, visit_type=expected_visit_type
    )
    assert request.from_webhook


def origin_save_webhook_receiver_invalid_request_test(
    forge_type: str,
    http_headers: Dict[str, Any],
    payload: Dict[str, Any],
    api_client,
):
    url = reverse(f"api-1-origin-save-webhook-{forge_type.lower()}")

    resp = check_api_post_response(
        api_client,
        url,
        status_code=400,
        data=payload,
        **django_http_headers(http_headers),
    )

    assert resp.data == {
        "exception": "BadInputExc",
        "reason": (
            f"POST request was not sent by a {forge_type} webhook "
            "and has not been processed."
        ),
    }


def origin_save_webhook_receiver_invalid_event_test(
    forge_type: str,
    http_headers: Dict[str, Any],
    payload: Dict[str, Any],
    api_client,
):
    url = reverse(f"api-1-origin-save-webhook-{forge_type.lower()}")

    resp = check_api_post_response(
        api_client,
        url,
        status_code=400,
        data=payload,
        **django_http_headers(http_headers),
    )

    assert resp.data == {
        "exception": "BadInputExc",
        "reason": (
            f"Event sent by {forge_type} webhook is not a push one, request has "
            "not been processed."
        ),
    }


def origin_save_webhook_receiver_invalid_content_type_test(
    forge_type: str,
    http_headers: Dict[str, Any],
    payload: Dict[str, Any],
    api_client,
):
    url = reverse(f"api-1-origin-save-webhook-{forge_type.lower()}")

    bad_content_type = "application/x-www-form-urlencoded"
    http_headers["Content-Type"] = bad_content_type

    resp = check_api_post_response(
        api_client,
        url,
        status_code=400,
        data=payload,
        **django_http_headers(http_headers),
    )

    assert resp.data == {
        "exception": "BadInputExc",
        "reason": (
            f"Invalid content type '{bad_content_type}' for the POST request sent by "
            f"{forge_type} webhook, it should be 'application/json'."
        ),
    }


def origin_save_webhook_receiver_no_repo_url_test(
    forge_type: str,
    http_headers: Dict[str, Any],
    payload: Dict[str, Any],
    api_client,
):
    url = reverse(f"api-1-origin-save-webhook-{forge_type.lower()}")

    resp = check_api_post_response(
        api_client,
        url,
        status_code=400,
        data=payload,
        **django_http_headers(http_headers),
    )

    assert resp.data == {
        "exception": "BadInputExc",
        "reason": (
            f"Repository URL could not be extracted from {forge_type} webhook payload."
        ),
    }


def origin_save_webhook_receiver_private_repo_test(
    forge_type: str,
    http_headers: Dict[str, Any],
    payload: Dict[str, Any],
    api_client,
    expected_origin_url: str,
):
    url = reverse(f"api-1-origin-save-webhook-{forge_type.lower()}")

    resp = check_api_post_response(
        api_client,
        url,
        status_code=400,
        data=payload,
        **django_http_headers(http_headers),
    )

    assert resp.data == {
        "exception": "BadInputExc",
        "reason": (
            f"Repository {expected_origin_url} is private and cannot be cloned "
            "without authentication."
        ),
    }


def origin_save_webhook_receiver_cooldown_requests_test(
    forge_type: str,
    http_headers: Dict[str, Any],
    payload: Dict[str, Any],
    api_client,
    swh_scheduler,
):
    url = reverse(f"api-1-origin-save-webhook-{forge_type.lower()}")

    resp = check_api_post_response(
        api_client,
        url,
        status_code=200,
        data=payload,
        **django_http_headers(http_headers),
    )

    # first webhook request should be executed immediately
    first_sor = get_save_origin_request(resp.data["id"])
    task = swh_scheduler.get_tasks([first_sor["loading_task_id"]])[0]
    assert task.next_run <= datetime.now(tz=timezone.utc)

    # simulate first task successful execution
    last_sor = SaveOriginRequest.objects.first()
    assert last_sor is not None
    last_sor.loading_task_status = SAVE_TASK_SUCCEEDED
    last_sor.visit_date = datetime.now(tz=timezone.utc)
    last_sor.visit_status = "full"
    last_sor.save()

    resp = check_api_post_response(
        api_client,
        url,
        status_code=200,
        data=payload,
        **django_http_headers(http_headers),
    )

    second_sor = get_save_origin_request(resp.data["id"])
    task = swh_scheduler.get_tasks([second_sor["loading_task_id"]])[0]

    # second webhook request in a row should delay loading task execution
    assert second_sor["id"] != first_sor["id"]
    assert task.next_run > datetime.now(tz=timezone.utc)
    assert task.next_run == iso8601.parse_date(
        first_sor["save_request_date"]
    ) + timedelta(minutes=WEBHOOK_REQUEST_COOLDOWN_INTERVAL)

    resp = check_api_post_response(
        api_client,
        url,
        status_code=200,
        data=payload,
        **django_http_headers(http_headers),
    )

    third_sor = get_save_origin_request(resp.data["id"])

    # third webhook request in a row should return the same request as previously
    assert third_sor["id"] == second_sor["id"]

    # modify request dates in order for the next request to be accepted
    # and executed immediately
    for sor_id in (first_sor["id"], second_sor["id"]):
        sor = SaveOriginRequest.objects.get(id=sor_id)
        sor.request_date = sor.request_date - timedelta(
            minutes=WEBHOOK_REQUEST_COOLDOWN_INTERVAL
        )
        sor.loading_task_status = SAVE_TASK_SUCCEEDED
        sor.visit_date = datetime.now(tz=timezone.utc)
        sor.visit_status = "full"
        sor.save()
    swh_scheduler.set_status_tasks(
        [task.id],
        next_run=datetime.now(tz=timezone.utc)
        - timedelta(minutes=WEBHOOK_REQUEST_COOLDOWN_INTERVAL),
    )

    resp = check_api_post_response(
        api_client,
        url,
        status_code=200,
        data=payload,
        **django_http_headers(http_headers),
    )

    fourth_sor = get_save_origin_request(resp.data["id"])
    task = swh_scheduler.get_tasks([fourth_sor["loading_task_id"]])[0]

    # fourth webhhok request should be executed immediately
    assert fourth_sor["id"] != third_sor["id"]
    assert task.next_run <= datetime.now(tz=timezone.utc)



# Copyright (C) 2022-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


import pytest
from requests import Response

from django.utils.html import escape

from swh.core.api import RemoteException
from swh.web.tests.django_asserts import assert_contains
from swh.web.tests.helpers import check_api_get_responses, check_html_get_response
from swh.web.utils import reverse
from swh.web.utils.exc import (
    BadInputExc,
    ForbiddenExc,
    Http404,
    Ratelimited,
    UnauthorizedExc,
    exceptions,
)


def _error_response():
    response = Response()
    response.status_code = 500
    return response


@pytest.mark.parametrize(
    "exception, expected_status",
    [
        (
            RemoteException(
                payload={
                    "type": "QueryCanceled",
                    "args": ["canceling statement due to statement timeout\n"],
                },
                response=_error_response(),
            ),
            500,
        ),
        (BadInputExc("bad input"), 400),
        (UnauthorizedExc("requires authentication"), 401),
        (ForbiddenExc("forbidden"), 403),
        (exceptions.PermissionDenied(), 403),
        (Http404("404"), 404),
        (Ratelimited(), 429),
    ],
)
def test_django_views_exception_handler(
    client, mocker, exception, expected_status, caplog, config_updater
):
    # simulate production settings
    config_updater({"debug": False})

    # make homepage return an error
    mocker.patch("swh.web.webapp.urls.origin_visit_types").side_effect = exception

    resp = check_html_get_response(
        client, "/", status_code=expected_status, template_used="error.html"
    )

    assert_contains(resp, escape(str(exception)), status_code=expected_status)

    if expected_status == 500:
        # error 500 should log exception
        assert caplog.records
        assert caplog.records[0].exc_info[1] == exception
    else:
        assert not caplog.records


@pytest.mark.parametrize(
    "exception, expected_status",
    [
        (
            RemoteException(
                payload={
                    "type": "QueryCanceled",
                    "args": ["canceling statement due to statement timeout\n"],
                },
                response=_error_response(),
            ),
            500,
        ),
        (BadInputExc("bad input"), 400),
        (UnauthorizedExc("requires authentication"), 401),
        (ForbiddenExc("forbidden"), 403),
    ],
)
def test_drf_views_exception_handler(
    api_client, mocker, exception, expected_status, caplog, config_updater
):
    # simulate production settings
    config_updater({"debug": False})

    # make stats api view return an error
    mocker.patch("swh.web.api.views.stat.archive.stat_counters").side_effect = exception

    resp = check_api_get_responses(
        api_client, reverse("api-1-stat-counters"), status_code=expected_status
    )

    assert resp.data["reason"] == str(exception)

    if expected_status == 500:
        assert caplog.records
        assert caplog.records[0].exc_info[1] == exception
    else:
        assert not caplog.records



# Copyright (C) 2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from pathlib import Path
import shutil

import pytest

from django.contrib.staticfiles import finders

from swh.web.tests.django_asserts import assert_contains, assert_not_contains
from swh.web.tests.helpers import check_html_get_response
from swh.web.utils import reverse

partner_name = "Example"
swh_extra_django_apps = [
    "swh.web.badges",
    "swh.web.jslicenses",
    "swh.web.vault",
]


@pytest.fixture
def mirror_config(tmp_path):
    static_path = os.path.join(tmp_path, "static")
    templates_path = os.path.join(tmp_path, "templates")
    os.makedirs(static_path)
    os.makedirs(templates_path)
    Path(os.path.join(templates_path, "mirror-homepage.html")).touch()
    Path(os.path.join(templates_path, "mirror-footer.html")).touch()

    return {
        "partner_name": partner_name,
        "partner_url": "https://example.org",
        "static_path": static_path,
        "partner_logo_static_path": "mirror-partner-logo.png",
        "templates_path": templates_path,
    }


@pytest.fixture(autouse=True)
def mirror_config_setter(mirror_config, config_updater):
    """Plug mirror config in django settings and reload the latters"""
    config_updater(
        {"mirror_config": mirror_config, "swh_extra_django_apps": swh_extra_django_apps}
    )


def test_page_titles_contain_mirror_partner_name(client):
    url = reverse("swh-web-homepage")
    response = client.get(url)

    mirror_info = f"{partner_name} Mirror of the Software Heritage archive"
    page_title = f"Welcome to the {mirror_info}"

    # welcome sentence in page title and navbar content
    assert_contains(response, page_title, count=2)

    url = reverse("browse-search")
    response = check_html_get_response(client, url, status_code=200)
    # mirror info in page title only
    assert_contains(response, mirror_info, count=1)


def test_pages_contain_mirror_logo(client):
    url = reverse("swh-web-homepage")
    response = check_html_get_response(client, url, status_code=200)
    assert_contains(response, "img/swh-mirror.png")


def test_pages_contain_mirror_partner_logo(client, mirror_config):
    swh_logo_path = finders.find("img/swh-logo.png")
    swh_partner_logo_path = os.path.join(
        mirror_config["static_path"], mirror_config["partner_logo_static_path"]
    )
    shutil.copyfile(swh_logo_path, swh_partner_logo_path)

    assert finders.find(mirror_config["partner_logo_static_path"]) is not None

    url = reverse("swh-web-homepage")
    response = check_html_get_response(client, url, status_code=200)
    assert_contains(response, mirror_config["partner_logo_static_path"])
    assert_contains(response, mirror_config["partner_url"])


def test_pages_contain_save_code_now_external_link(client, origin):
    save_code_now_link = "https://archive.softwareheritage.org/save"
    url = reverse("swh-web-homepage")
    response = check_html_get_response(client, url, status_code=200)
    assert_contains(response, save_code_now_link)

    url = reverse("browse-origin-directory", query_params={"origin_url": origin["url"]})
    response = check_html_get_response(client, url, status_code=200)
    assert_contains(response, save_code_now_link + f"?origin_url={origin['url']}")


def test_admin_menu_is_not_available(client, admin_user):
    client.force_login(admin_user)
    url = reverse("swh-web-homepage")
    response = check_html_get_response(client, url, status_code=200)
    assert_not_contains(response, '<li class="nav-header">Administration</li>')


def test_pages_contain_add_forge_now_external_link(client):
    add_forge_now_link = (
        "https://archive.softwareheritage.org/add-forge/request/create/"
    )
    url = reverse("swh-web-homepage")
    response = check_html_get_response(client, url, status_code=200)
    assert_contains(response, add_forge_now_link)


def test_mirror_custom_homepage_and_footer(client, mirror_config):
    mirror_homepage_html = "<h4>Mirror section</h4>\n"
    mirror_footer_html = "<span>Mirror footer</span>\n"
    templates_path = mirror_config["templates_path"]
    with open(
        os.path.join(templates_path, "mirror-homepage.html"), "w"
    ) as mirror_homepage:
        mirror_homepage.write(mirror_homepage_html)

    with open(os.path.join(templates_path, "mirror-footer.html"), "w") as mirror_footer:
        mirror_footer.write(mirror_footer_html)
    url = reverse("swh-web-homepage")
    response = check_html_get_response(client, url, status_code=200)
    assert_contains(response, mirror_homepage_html)
    assert_contains(response, mirror_footer_html)



# Copyright (C) 2021-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import random

from swh.model.hashutil import hash_to_bytes
from swh.model.swhids import CoreSWHID, ObjectType, QualifiedSWHID
from swh.web.tests.django_asserts import assert_contains
from swh.web.tests.helpers import check_html_get_response
from swh.web.utils import reverse


def test_content_swhid_iframe(client, content_swhid):
    url = reverse("browse-swhid-iframe", url_args={"swhid": str(content_swhid)})
    check_html_get_response(
        client, url, status_code=200, template_used="browse-iframe.html"
    )


def test_content_core_swhid_iframe(client, content_swhid):
    content_core_swhid = CoreSWHID(
        object_type=content_swhid.object_type, object_id=content_swhid.object_id
    )
    url = reverse("browse-swhid-iframe", url_args={"swhid": str(content_core_swhid)})
    check_html_get_response(
        client, url, status_code=200, template_used="browse-iframe.html"
    )


def test_directory_swhid_iframe(client, directory_swhid):
    url = reverse("browse-swhid-iframe", url_args={"swhid": str(directory_swhid)})
    check_html_get_response(
        client, url, status_code=200, template_used="browse-iframe.html"
    )


def test_directory_core_swhid_iframe(client, directory_swhid):
    directory_core_swhid = CoreSWHID(
        object_type=directory_swhid.object_type, object_id=directory_swhid.object_id
    )
    url = reverse("browse-swhid-iframe", url_args={"swhid": str(directory_core_swhid)})
    check_html_get_response(
        client, url, status_code=200, template_used="browse-iframe.html"
    )


def test_iframe_unsupported_object(client, revision_swhid):
    url = reverse("browse-swhid-iframe", url_args={"swhid": str(revision_swhid)})
    check_html_get_response(
        client, url, status_code=400, template_used="browse-iframe.html"
    )


def test_iframe_object_not_found(client, unknown_directory):
    swhid = CoreSWHID(
        object_type=ObjectType.DIRECTORY, object_id=hash_to_bytes(unknown_directory)
    )
    url = reverse("browse-swhid-iframe", url_args={"swhid": str(swhid)})
    check_html_get_response(
        client, url, status_code=404, template_used="browse-iframe.html"
    )


def test_swhid_iframe_unknown_error(client, mocker, content_swhid):
    mocker.patch("swh.web.browse.views.iframe.get_snapshot_context").side_effect = (
        Exception("Error")
    )
    url = reverse("browse-swhid-iframe", url_args={"swhid": str(content_swhid)})
    check_html_get_response(
        client, url, status_code=500, template_used="browse-iframe.html"
    )


def test_iframe_directory_no_snapshot_context(
    client, archive_data, directory_with_subdirs
):
    dir_content = archive_data.directory_ls(directory_with_subdirs)
    subdir = random.choice([e for e in dir_content if e["type"] == "dir"])
    path = f"/{subdir['name']}/"

    root_swhid = CoreSWHID(
        object_type=ObjectType.DIRECTORY,
        object_id=hash_to_bytes(directory_with_subdirs),
    )
    swhid = CoreSWHID(
        object_type=ObjectType.DIRECTORY, object_id=hash_to_bytes(subdir["target"])
    )
    qualified_swhid = QualifiedSWHID(
        object_type=ObjectType.DIRECTORY,
        object_id=hash_to_bytes(subdir["target"]),
        anchor=root_swhid,
        path=path,
    )

    url = reverse(
        "browse-swhid-iframe",
        url_args={"swhid": f"{str(swhid)};path={path}"},
        query_params={"focus_swhid": str(root_swhid)},
    )
    resp = check_html_get_response(
        client, url, status_code=200, template_used="browse-iframe.html"
    )

    archive_url = reverse("browse-swhid", url_args={"swhid": str(qualified_swhid)})

    assert_contains(resp, archive_url)


def test_iframe_legacy_url_redirection(client, directory_swhid):
    directory_core_swhid = CoreSWHID(
        object_type=directory_swhid.object_type, object_id=directory_swhid.object_id
    )
    url = reverse(
        "browse-swhid-iframe-legacy", url_args={"swhid": str(directory_core_swhid)}
    )
    resp = check_html_get_response(
        client,
        url,
        status_code=302,
    )

    assert resp["Location"] == reverse(
        "browse-swhid-iframe", url_args={"swhid": str(directory_core_swhid)}
    )


def test_swhid_iframe_invalid_focus_swhid(client, content_swhid):
    url = reverse(
        "browse-swhid-iframe",
        url_args={"swhid": str(content_swhid)},
        query_params={"focus_swhid": "foo"},
    )
    check_html_get_response(
        client, url, status_code=200, template_used="browse-iframe.html"
    )



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest


@pytest.fixture
def make_masked_object_exception():
    def _masked_object_exception(swhid_str: str):
        from uuid import UUID

        from swh.model.swhids import ExtendedSWHID
        from swh.storage.exc import MaskedObjectException
        from swh.storage.proxies.masking.db import MaskedState, MaskedStatus

        swhid = ExtendedSWHID.from_string(swhid_str)
        status = MaskedStatus(
            state=MaskedState.DECISION_PENDING,
            request=UUID("da785a27-7e59-4a35-b82a-a5ae3714407c"),
        )
        return MaskedObjectException({swhid: [status]})

    return _masked_object_exception



# Generated by Django 2.2.24 on 2022-03-21 16:37

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("swh_web_add_forge_now", "0002_authorized_null_comment"),
    ]

    operations = [
        migrations.AddField(
            model_name="request",
            name="submitter_forward_username",
            field=models.BooleanField(default=False),
        ),
    ]



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import random

from swh.web.api.utils import enrich_directory_entry
from swh.web.tests.data import random_sha1
from swh.web.tests.helpers import check_api_get_responses, check_http_get_response
from swh.web.utils import reverse


def test_api_directory(api_client, archive_data, directory):
    url = reverse("api-1-directory", url_args={"sha1_git": directory})
    rv = check_api_get_responses(api_client, url, status_code=200)

    dir_content = list(archive_data.directory_ls(directory))
    expected_data = list(
        map(enrich_directory_entry, dir_content, [rv.wsgi_request] * len(dir_content))
    )

    assert rv.data == expected_data


def test_api_directory_not_found(api_client):
    unknown_directory_ = random_sha1()

    url = reverse("api-1-directory", url_args={"sha1_git": unknown_directory_})
    rv = check_api_get_responses(api_client, url, status_code=404)
    assert rv.data == {
        "exception": "NotFoundExc",
        "reason": "Directory with sha1_git %s not found" % unknown_directory_,
    }


def test_api_directory_with_path_found(api_client, archive_data, directory):
    directory_content = archive_data.directory_ls(directory)
    dir_entry = random.choice(directory_content)

    url = reverse(
        "api-1-directory", url_args={"sha1_git": directory, "path": dir_entry["name"]}
    )
    rv = check_api_get_responses(api_client, url, status_code=200)
    assert rv.data == enrich_directory_entry(dir_entry, rv.wsgi_request)


def test_api_directory_with_path_not_found(api_client, directory):
    path = "some/path/to/nonexistent/dir/"
    url = reverse("api-1-directory", url_args={"sha1_git": directory, "path": path})
    rv = check_api_get_responses(api_client, url, status_code=404)
    reason = (
        f"Directory entry with path {path} from root directory {directory} not found"
    )
    assert rv.data == {
        "exception": "NotFoundExc",
        "reason": reason,
    }


def test_api_directory_uppercase(api_client, directory):
    url = reverse(
        "api-1-directory-uppercase-checksum", url_args={"sha1_git": directory.upper()}
    )

    resp = check_http_get_response(api_client, url, status_code=302)

    redirect_url = reverse("api-1-directory", url_args={"sha1_git": directory})

    assert resp["location"] == redirect_url



# Copyright (C) 2019-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime, timedelta, timezone
from typing import Optional
import uuid

import iso8601
import pytest

from swh.scheduler.utils import create_oneshot_task
from swh.web.save_code_now.models import (
    SAVE_REQUEST_ACCEPTED,
    SAVE_TASK_FAILED,
    SAVE_TASK_RUNNING,
    SAVE_TASK_SUCCEEDED,
    VISIT_STATUS_CREATED,
    VISIT_STATUS_FULL,
    VISIT_STATUS_ONGOING,
    VISIT_STATUS_PARTIAL,
    SaveOriginRequest,
)
from swh.web.save_code_now.origin_save import (
    _check_origin_exists,
    _check_visit_type_savable,
    _visit_type_task,
    _visit_type_task_privileged,
    get_savable_visit_types,
    get_save_origin_requests,
    get_save_origin_task_info,
    origin_exists,
    refresh_save_origin_request_statuses,
)
from swh.web.utils.exc import BadInputExc
from swh.web.utils.typing import (
    OriginExistenceCheckInfo,
    OriginVisitInfo,
    SaveOriginRequestInfo,
)

_origin_url = "https://gitlab.com/inkscape/inkscape"
_visit_type = "git"


@pytest.mark.django_db
def test_get_save_origin_archived_task_info(swh_scheduler):
    _get_save_origin_task_info_test(swh_scheduler, task_archived=True)


@pytest.mark.django_db
def test_get_save_origin_task_info(swh_scheduler):
    _get_save_origin_task_info_test(swh_scheduler)


def _fill_scheduler_db(
    swh_scheduler,
    task_status="completed",
    task_run_status="eventful",
    task_archived=False,
    visit_started_date=None,
):
    task = task_run = None
    if not task_archived:
        task = swh_scheduler.create_tasks(
            [create_oneshot_task("load-git", repo_url=_origin_url)]
        )[0]
        backend_id = str(uuid.uuid4())

        if task_status != "next_run_not_scheduled":
            swh_scheduler.schedule_task_run(task.id, backend_id)

        if task_run_status is not None:
            swh_scheduler.start_task_run(backend_id)
            task_run = swh_scheduler.end_task_run(backend_id, task_run_status)
    return task, task_run


@pytest.mark.parametrize(
    "wrong_type,privileged_user",
    [
        ("dummy", True),
        ("dumb", False),
        ("archives", False),  # when no privilege, this is rejected
    ],
)
def test_check_visit_type_savable(wrong_type, privileged_user, swh_scheduler):
    swh_scheduler.add_load_archive_task_type()

    with pytest.raises(BadInputExc, match="Allowed types"):
        _check_visit_type_savable(wrong_type, privileged_user)

    # when privileged_user, the following is accepted though
    _check_visit_type_savable("archives", True)


def test_get_savable_visit_types(swh_scheduler):
    swh_scheduler.add_load_archive_task_type()

    default_list = list(_visit_type_task.keys())

    assert set(get_savable_visit_types()) == set(default_list)

    privileged_list = default_list.copy()
    privileged_list += list(_visit_type_task_privileged.keys())

    assert set(get_savable_visit_types(privileged_user=True)) == set(privileged_list)


def _get_save_origin_task_info_test(
    swh_scheduler,
    task_archived=False,
):
    task, task_run = _fill_scheduler_db(swh_scheduler, task_archived=task_archived)

    sor = SaveOriginRequest.objects.create(
        request_date=datetime.now(tz=timezone.utc),
        visit_type=_visit_type,
        origin_url="https://gitlab.com/inkscape/inkscape",
        status=SAVE_REQUEST_ACCEPTED,
        visit_date=datetime.now(tz=timezone.utc) + timedelta(hours=1),
        loading_task_id=task.id if task else 0,
    )

    sor_task_info = get_save_origin_task_info(sor.id)

    expected_result = (
        {
            "type": task.type,
            "arguments": task.arguments.to_dict(),
            "id": task.id,
            "backend_id": task_run.backend_id,
            "scheduled": task_run.scheduled,
            "started": task_run.started,
            "ended": task_run.ended,
            "status": task_run.status,
            "visit_status": sor.visit_status,
            "metadata": {},
        }
        if not task_archived
        else {}
    )

    if "message" in expected_result:
        message = ""
        message_lines = expected_result["message"].split("\n")
        for line in message_lines:
            if line.startswith("Traceback"):
                break
            message += f"{line}\n"
        message += message_lines[-1]
        expected_result["message"] = message

    assert sor_task_info == expected_result


@pytest.mark.django_db
def test_get_save_origin_requests_find_visit_date(mocker, swh_scheduler):
    # create a save request
    task, _ = _fill_scheduler_db(swh_scheduler)

    SaveOriginRequest.objects.create(
        request_date=datetime.now(tz=timezone.utc),
        visit_type=_visit_type,
        origin_url=_origin_url,
        status=SAVE_REQUEST_ACCEPTED,
        visit_date=None,
        loading_task_id=task.id,
    )

    # mock scheduler and archive
    mock_archive = mocker.patch("swh.web.save_code_now.origin_save.archive")
    mock_archive.lookup_origin.return_value = {"url": _origin_url}
    # create a visit for the save request
    visit_date = datetime.now(tz=timezone.utc).isoformat()
    visit_info = OriginVisitInfo(
        date=visit_date,
        formatted_date="",
        metadata={},
        origin=_origin_url,
        snapshot="",
        status=VISIT_STATUS_FULL,
        type=_visit_type,
        url="",
        visit=34,
    )
    mock_archive.origin_visit_find_by_date.return_value = visit_info

    # check visit date has been correctly found
    sors = get_save_origin_requests(_visit_type, _origin_url)
    assert len(sors) == 1
    assert sors[0]["save_task_status"] == SAVE_TASK_SUCCEEDED
    assert sors[0]["visit_date"] == visit_date
    mock_archive.origin_visit_find_by_date.assert_called()


def _get_save_origin_requests(
    mocker,
    swh_scheduler,
    load_status,
    visit_status,
    request_date: Optional[datetime] = None,
):
    """Wrapper around the get_origin_save_origin_request call."""

    task, task_run = _fill_scheduler_db(
        swh_scheduler, task_status="next_run_scheduled", task_run_status=load_status
    )

    assert task is not None
    SaveOriginRequest.objects.create(
        request_date=datetime.now(tz=timezone.utc),
        visit_type=_visit_type,
        visit_status=visit_status,
        origin_url=_origin_url,
        status=SAVE_REQUEST_ACCEPTED,
        visit_date=None,
        loading_task_id=task.id,
    )

    # mock scheduler and archives
    mock_archive = mocker.patch("swh.web.save_code_now.origin_save.archive")
    mock_archive.lookup_origin.return_value = {"url": _origin_url}
    # create a visit for the save request with status created
    visit_date = datetime.now(tz=timezone.utc).isoformat()
    visit_info = OriginVisitInfo(
        date=visit_date,
        formatted_date="",
        metadata={},
        origin=_origin_url,
        snapshot="",  # make mypy happy
        status=visit_status,
        type=_visit_type,
        url="",
        visit=34,
    )
    mock_archive.origin_visit_find_by_date.return_value = visit_info

    sors = get_save_origin_requests(_visit_type, _origin_url)
    mock_archive.origin_visit_find_by_date.assert_called()

    return sors


@pytest.mark.parametrize("visit_date", [None, "some-date"])
def test_from_save_origin_request_to_save_request_info_dict(visit_date, snapshot):
    """Ensure save request to json serializable dict is fine"""
    request_date = datetime.now(tz=timezone.utc)
    _visit_date = request_date + timedelta(minutes=5) if visit_date else None
    request_date = datetime.now(tz=timezone.utc)
    note = "request succeeded"
    snapshot_swhid = f"swh:1:snp:{snapshot}"
    sor = SaveOriginRequest(
        request_date=request_date,
        visit_type=_visit_type,
        visit_status=VISIT_STATUS_FULL,
        origin_url=_origin_url,
        status=SAVE_REQUEST_ACCEPTED,
        loading_task_status=None,
        visit_date=_visit_date,
        loading_task_id=1,
        note=note,
        snapshot_swhid=snapshot_swhid,
    )

    assert sor.to_dict() == SaveOriginRequestInfo(
        id=sor.id,
        origin_url=sor.origin_url,
        visit_type=sor.visit_type,
        save_request_date=sor.request_date.isoformat(),
        save_request_status=sor.status,
        save_task_status=sor.loading_task_status,
        visit_status=sor.visit_status,
        visit_date=_visit_date.isoformat() if _visit_date else None,
        loading_task_id=sor.loading_task_id,
        note=note,
        from_webhook=False,
        webhook_origin="",
        snapshot_swhid=snapshot_swhid,
    )


def test__check_origin_exists_404(requests_mock):
    url_ko = "https://example.org/some-inexistant-url"
    requests_mock.head(url_ko, status_code=404)

    with pytest.raises(BadInputExc, match="not exist"):
        _check_origin_exists(url_ko)


def test__check_origin_exists_200(requests_mock):
    url = "https://example.org/url"
    requests_mock.head(url, status_code=200)

    # passes the check
    actual_metadata = _check_origin_exists(url)

    # and we actually may have retrieved some metadata on the origin
    assert actual_metadata == origin_exists(url)


def test_origin_exists_404(requests_mock):
    """Origin which does not exist should be reported as inexistent"""
    url_ko = "https://example.org/some-inexistant-url"
    requests_mock.head(url_ko, status_code=404)

    actual_result = origin_exists(url_ko)
    assert actual_result == OriginExistenceCheckInfo(
        origin_url=url_ko,
        exists=False,
        last_modified=None,
        content_length=None,
    )


def test_origin_exists_200_no_data(requests_mock):
    """Existing origin should be reported as such (no extra information)"""
    url = "http://example.org/real-url"
    requests_mock.head(
        url,
        status_code=200,
    )

    actual_result = origin_exists(url)
    assert actual_result == OriginExistenceCheckInfo(
        origin_url=url,
        exists=True,
        last_modified=None,
        content_length=None,
    )


def test_origin_exists_200_with_data(requests_mock):
    """Existing origin should be reported as such (+ extra information)"""
    url = "http://example.org/real-url"
    requests_mock.head(
        url,
        status_code=200,
        headers={
            "content-length": "10",
            "last-modified": "Sun, 21 Aug 2011 16:26:32 GMT",
        },
    )

    actual_result = origin_exists(url)
    assert actual_result == OriginExistenceCheckInfo(
        origin_url=url,
        exists=True,
        content_length=10,
        last_modified="2011-08-21T16:26:32",
    )


def test_origin_exists_internet_archive(requests_mock):
    """Edge case where an artifact URL to check existence is hosted on the
    Internet Archive"""
    url = (
        "https://web.archive.org/web/20100705043309/"
        "http://www.cs.unm.edu/~mccune/old-ftp/eqp-09e.tar.gz"
    )
    redirect_url = (
        "https://web.archive.org/web/20100610004108/"
        "http://www.cs.unm.edu/~mccune/old-ftp/eqp-09e.tar.gz"
    )
    requests_mock.head(
        url,
        status_code=302,
        headers={
            "Location": redirect_url,
        },
    )
    requests_mock.head(
        redirect_url,
        status_code=200,
        headers={
            "X-Archive-Orig-Last-Modified": "Tue, 12 May 2009 22:09:43 GMT",
            "X-Archive-Orig-Content-Length": "121421",
        },
    )

    actual_result = origin_exists(url)
    assert actual_result == OriginExistenceCheckInfo(
        origin_url=url,
        exists=True,
        content_length=121421,
        last_modified="2009-05-12T22:09:43",
    )


def test_origin_exists_200_with_data_unexpected_date_format(requests_mock):
    """Existing origin should be ok, unexpected last modif time result in no time"""
    url = "http://example.org/real-url2"
    # this is parsable but not as expected
    unexpected_format_date = "Sun, 21 Aug 2021 16:26:32"
    requests_mock.head(
        url,
        status_code=200,
        headers={
            "last-modified": unexpected_format_date,
        },
    )

    actual_result = origin_exists(url)
    # so the resulting date is None
    assert actual_result == OriginExistenceCheckInfo(
        origin_url=url,
        exists=True,
        content_length=None,
        last_modified=None,
    )


@pytest.mark.django_db
@pytest.mark.parametrize(
    "visit_status",
    [
        VISIT_STATUS_CREATED,
        VISIT_STATUS_ONGOING,
    ],
)
def test_get_save_origin_requests_no_visit_date_found(
    mocker, swh_scheduler, visit_status
):
    """Uneventful visits with failed visit status are marked as failed"""
    sors = _get_save_origin_requests(
        mocker,
        swh_scheduler,
        load_status="started",
        visit_status=visit_status,
    )
    # check no visit date has been found
    assert len(sors) == 1
    assert sors[0]["save_task_status"] == SAVE_TASK_RUNNING
    assert sors[0]["visit_date"] is not None
    assert sors[0]["visit_status"] == visit_status


@pytest.mark.django_db
@pytest.mark.parametrize(
    "visit_status",
    [
        "not_found",
        "failed",
    ],
)
def test_get_save_origin_requests_no_failed_status_override(
    mocker, swh_scheduler, visit_status
):
    """Uneventful visits with failed statuses (failed, not found) are marked as failed"""
    sors = _get_save_origin_requests(
        mocker, swh_scheduler, load_status="uneventful", visit_status=visit_status
    )
    assert len(sors) == 1

    assert sors[0]["save_task_status"] == SAVE_TASK_FAILED
    visit_date = sors[0]["visit_date"]
    assert visit_date is not None

    sors = get_save_origin_requests(_visit_type, _origin_url)
    assert len(sors) == 1
    assert sors[0]["save_task_status"] == SAVE_TASK_FAILED
    assert sors[0]["visit_status"] == visit_status


@pytest.mark.django_db
@pytest.mark.parametrize(
    "load_status,visit_status",
    [
        ("eventful", VISIT_STATUS_FULL),
        ("eventful", VISIT_STATUS_PARTIAL),
        ("uneventful", VISIT_STATUS_PARTIAL),
    ],
)
def test_get_visit_info_for_save_request_succeeded(
    mocker, swh_scheduler, load_status, visit_status
):
    """Nominal scenario, below 30 days, returns something"""
    sors = _get_save_origin_requests(
        mocker, swh_scheduler, load_status=load_status, visit_status=visit_status
    )
    assert len(sors) == 1

    assert sors[0]["save_task_status"] == SAVE_TASK_SUCCEEDED
    assert sors[0]["visit_date"] is not None
    assert sors[0]["visit_status"] == visit_status

    sors = get_save_origin_requests(_visit_type, _origin_url)
    assert sors[0]["save_task_status"] == SAVE_TASK_SUCCEEDED
    assert sors[0]["visit_status"] == visit_status


@pytest.mark.django_db
@pytest.mark.parametrize(
    "load_status",
    [
        "eventful",
        "uneventful",
    ],
)
def test_get_visit_info_incomplete_visit_still_successful(
    mocker, swh_scheduler, load_status
):
    """Incomplete visit information, yet the task is updated partially"""
    sors = _get_save_origin_requests(
        mocker,
        swh_scheduler,
        load_status=load_status,
        visit_status=None,
    )
    assert len(sors) == 1

    assert sors[0]["save_task_status"] == SAVE_TASK_SUCCEEDED
    # As the entry is missing the following information though
    assert sors[0]["visit_date"] is not None
    assert sors[0]["visit_status"] is None

    # It's still detected as to be updated by the refresh routine
    sors = refresh_save_origin_request_statuses()
    assert len(sors) == 1

    assert sors[0]["save_task_status"] == SAVE_TASK_SUCCEEDED
    assert sors[0]["visit_date"] is not None
    assert sors[0]["visit_status"] is None


@pytest.mark.django_db
def test_refresh_in_progress_save_request_statuses(mocker, swh_scheduler):
    """Refresh a pending save origins requests and update if the status changes"""
    date_now = datetime.now(tz=timezone.utc)
    date_pivot = date_now - timedelta(days=30)
    visit_started_date = date_now - timedelta(minutes=1)

    task, task_run = _fill_scheduler_db(
        swh_scheduler,
        task_status="next_run_scheduled",
        task_run_status="started",
    )
    assert task is not None

    # returned visit status
    SaveOriginRequest.objects.create(
        request_date=datetime.now(tz=timezone.utc),
        visit_type=_visit_type,
        visit_status=VISIT_STATUS_CREATED,
        origin_url=_origin_url,
        status=SAVE_REQUEST_ACCEPTED,
        visit_date=None,
        loading_task_id=task.id,
    )

    # mock scheduler and archives
    mock_archive = mocker.patch("swh.web.save_code_now.origin_save.archive")
    mock_archive.lookup_origin.return_value = {"url": _origin_url}
    # create a visit for the save request with status created
    visit_date = datetime.now(tz=timezone.utc).isoformat()
    visit_info = OriginVisitInfo(
        date=visit_date,
        formatted_date="",
        metadata={},
        origin=_origin_url,
        snapshot="",  # make mypy happy
        status=VISIT_STATUS_CREATED,
        type=_visit_type,
        url="",
        visit=34,
    )
    mock_archive.origin_visit_find_by_date.return_value = visit_info

    # make the scheduler return a running event
    _fill_scheduler_db(
        swh_scheduler,
        task_status="next_run_scheduled",
        task_run_status="started",
        visit_started_date=visit_started_date,
    )

    # The visit is detected but still running
    sors = refresh_save_origin_request_statuses()

    assert (
        mock_archive.origin_visit_find_by_date.called
        and mock_archive.origin_visit_find_by_date.call_count == 1
    )

    assert len(sors) == 1

    for sor in sors:
        assert iso8601.parse_date(sor["save_request_date"]) >= date_pivot
        # The status is updated
        assert sor["save_task_status"] == SAVE_TASK_RUNNING
        # but the following entries are missing so it's not updated
        assert sor["visit_date"] is not None
        assert sor["visit_status"] == VISIT_STATUS_CREATED

    # make the visit status completed
    # make the scheduler return a running event
    _fill_scheduler_db(
        swh_scheduler,
        task_status="completed",
        task_run_status="eventful",
        visit_started_date=visit_started_date,
    )

    # This time around, the origin returned will have all required information updated
    # (visit date and visit status in final state)
    visit_date = datetime.now(tz=timezone.utc).isoformat()
    visit_info.update({"date": visit_date, "status": VISIT_STATUS_FULL})
    mock_archive.origin_visit_find_by_date.return_value = visit_info

    # Detected entry, this time it should be updated
    sors = refresh_save_origin_request_statuses()
    assert len(sors) == 1

    assert (
        mock_archive.origin_visit_find_by_date.called
        and mock_archive.origin_visit_find_by_date.call_count == 3
    )

    for sor in sors:
        assert iso8601.parse_date(sor["save_request_date"]) >= date_pivot
        # as it turns out, in this test, this won't update anything as no new status got
        # returned by the scheduler
        assert sor["save_task_status"] == SAVE_TASK_SUCCEEDED
        assert sor["visit_date"] == visit_date
        assert sor["visit_status"] == VISIT_STATUS_FULL

    # Once in final state, a sor should not be updated anymore
    sors = refresh_save_origin_request_statuses()
    assert len(sors) == 0


@pytest.mark.django_db
def test_refresh_save_request_statuses(mocker, swh_scheduler):
    """Refresh filters save origins requests and update if changes"""
    date_now = datetime.now(tz=timezone.utc)
    date_pivot = date_now - timedelta(days=30)
    # returned visit status

    task, task_run = _fill_scheduler_db(
        swh_scheduler,
        task_status="next_run_scheduled",
        task_run_status="started",
    )
    assert task is not None
    SaveOriginRequest.objects.create(
        request_date=datetime.now(tz=timezone.utc),
        visit_type=_visit_type,
        visit_status=None,
        origin_url=_origin_url,
        status=SAVE_REQUEST_ACCEPTED,
        visit_date=None,
        loading_task_id=task.id,
    )

    # mock scheduler and archives
    mock_archive = mocker.patch("swh.web.save_code_now.origin_save.archive")
    mock_archive.lookup_origin.return_value = {"url": _origin_url}
    # create a visit for the save request with status created
    visit_date = datetime.now(tz=timezone.utc).isoformat()
    visit_info = OriginVisitInfo(
        date=visit_date,
        formatted_date="",
        metadata={},
        origin=_origin_url,
        snapshot="",  # make mypy happy
        status=VISIT_STATUS_CREATED,
        type=_visit_type,
        url="",
        visit=34,
    )
    mock_archive.origin_visit_find_by_date.return_value = visit_info

    # no changes so refresh does detect the entry but does nothing
    sors = refresh_save_origin_request_statuses()
    assert len(sors) == 1

    for sor in sors:
        assert iso8601.parse_date(sor["save_request_date"]) >= date_pivot
        # as it turns out, in this test, this won't update anything as no new status got
        # returned by the scheduler
        assert sor["save_task_status"] == SAVE_TASK_RUNNING
        # Information is empty
        assert sor["visit_date"] == visit_date
        assert sor["visit_status"] == VISIT_STATUS_CREATED

    # A save code now entry is detected for update, but as nothing changes, the entry
    # remains in the same state
    sors = refresh_save_origin_request_statuses()
    assert len(sors) == 1

    for sor in sors:
        assert iso8601.parse_date(sor["save_request_date"]) >= date_pivot
        # Status is not updated as no new information is available on the visit status
        # and the task status has not moved
        assert sor["save_task_status"] == SAVE_TASK_RUNNING
        # Information is empty
        assert sor["visit_date"] == visit_date
        assert sor["visit_status"] == VISIT_STATUS_CREATED

    # This time around, the origin returned will have all information updated
    # create a visit for the save request with status created
    visit_date = datetime.now(tz=timezone.utc).isoformat()
    visit_info = OriginVisitInfo(
        date=visit_date,
        formatted_date="",
        metadata={},
        origin=_origin_url,
        snapshot="",  # make mypy happy
        status=VISIT_STATUS_FULL,
        type=_visit_type,
        url="",
        visit=34,
    )
    mock_archive.origin_visit_find_by_date.return_value = visit_info

    # Detected entry, this time it should be updated
    sors = refresh_save_origin_request_statuses()
    assert len(sors) == 1

    for sor in sors:
        assert iso8601.parse_date(sor["save_request_date"]) >= date_pivot
        # as it turns out, in this test, this won't update anything as no new status got
        # returned by the scheduler
        assert sor["save_task_status"] == SAVE_TASK_SUCCEEDED
        assert sor["visit_date"] == visit_date
        assert sor["visit_status"] == VISIT_STATUS_FULL

    # This time, nothing left to update
    sors = refresh_save_origin_request_statuses()
    assert len(sors) == 0


@pytest.mark.django_db
def test_refresh_save_request_when_worker_lost(mocker, swh_scheduler):
    """Check save request status is set to failed when the celery worker
    exited prematurely."""

    # simulate a running save request
    task, task_run = _fill_scheduler_db(
        swh_scheduler,
        task_status="next_run_scheduled",
        task_run_status="started",
    )
    assert task is not None
    SaveOriginRequest.objects.create(
        request_date=datetime.now(tz=timezone.utc),
        visit_type=_visit_type,
        visit_status=VISIT_STATUS_ONGOING,
        origin_url=_origin_url,
        status=SAVE_REQUEST_ACCEPTED,
        visit_date=None,
        loading_task_id=task.id,
    )

    # mock scheduler and archives
    mock_archive = mocker.patch("swh.web.save_code_now.origin_save.archive")
    mock_archive.lookup_origin.return_value = {"url": _origin_url}
    # create a visit for the save request with status 'running'
    visit_date = datetime.now(tz=timezone.utc).isoformat()
    visit_info = OriginVisitInfo(
        date=visit_date,
        formatted_date="",
        metadata={},
        origin=_origin_url,
        snapshot="",  # make mypy happy
        status=VISIT_STATUS_ONGOING,
        type=_visit_type,
        url="",
        visit=34,
    )
    mock_archive.origin_visit_find_by_date.return_value = visit_info

    # save task should be marked as running
    sors = refresh_save_origin_request_statuses()
    assert sors[0]["save_task_status"] == SAVE_TASK_RUNNING

    # simulate lost of celery worker
    swh_scheduler.end_task_run(task_run.backend_id, "failed")

    # save task should be marked as failed
    sors = refresh_save_origin_request_statuses()
    assert sors[0]["save_task_status"] == SAVE_TASK_FAILED



# Copyright (C) 2022-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import copy
import datetime
import threading
import time
from typing import Dict
from urllib.parse import urlencode, urlparse

import iso8601
import pytest

from swh.web.add_forge_now.models import Request, RequestHistory
from swh.web.config import get_config
from swh.web.inbound_email.utils import get_address_for_pk
from swh.web.tests.helpers import (
    check_api_get_responses,
    check_api_post_response,
    check_http_get_response,
    check_http_post_response,
)
from swh.web.utils import reverse


@pytest.mark.django_db
def test_add_forge_request_create_anonymous_user(api_client):
    url = reverse("api-1-add-forge-request-create")
    check_api_post_response(api_client, url, status_code=403)


@pytest.mark.django_db
def test_add_forge_request_create_empty(api_client, regular_user):
    api_client.force_login(regular_user)
    url = reverse("api-1-add-forge-request-create")
    resp = check_api_post_response(api_client, url, status_code=400)
    assert '"forge_type"' in resp.data["reason"]


ADD_FORGE_DATA_FORGE1: Dict = {
    "forge_type": "gitlab",
    "forge_url": "https://gitlab.example.org",
    "forge_contact_email": "admin@gitlab.example.org",
    "forge_contact_name": "gitlab.example.org admin",
    "forge_contact_comment": "user marked as owner in forge members",
    "submitter_forward_username": True,
}

ADD_FORGE_DATA_FORGE2: Dict = {
    "forge_type": "gitea",
    "forge_url": "https://gitea.example.org",
    "forge_contact_email": "admin@gitea.example.org",
    "forge_contact_name": "gitea.example.org admin",
    "forge_contact_comment": "user marked as owner in forge members",
    "submitter_forward_username": True,
}

ADD_FORGE_DATA_FORGE3: Dict = {
    "forge_type": "heptapod",
    "forge_url": "https://heptapod.host/",
    "forge_contact_email": "admin@example.org",
    "forge_contact_name": "heptapod admin",
    "forge_contact_comment": "",  # authorized empty or null comment
    "submitter_forward_username": False,
}

ADD_FORGE_DATA_FORGE4: Dict = {
    **ADD_FORGE_DATA_FORGE3,
    "forge_url": "https://heptapod2.host/",
    "submitter_forward_username": "on",
}

ADD_FORGE_DATA_FORGE5: Dict = {
    **ADD_FORGE_DATA_FORGE3,
    "forge_url": "https://heptapod3.host/",
    "submitter_forward_username": "off",
}


def inbound_email_for_pk(pk: int) -> str:
    """Check that the inbound email matches the one expected for the given pk"""

    base_address = get_config()["add_forge_now"]["email_address"]
    return get_address_for_pk(
        salt="swh_web_add_forge_now", base_address=base_address, pk=pk
    )


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize(
    "add_forge_data",
    [
        ADD_FORGE_DATA_FORGE1,
        ADD_FORGE_DATA_FORGE2,
        ADD_FORGE_DATA_FORGE3,
        ADD_FORGE_DATA_FORGE4,
    ],
)
def test_add_forge_request_create_success_post(
    api_client, regular_user, add_forge_data
):
    api_client.force_login(regular_user)
    url = reverse("api-1-add-forge-request-create")

    date_before = datetime.datetime.now(tz=datetime.timezone.utc)

    resp = check_api_post_response(
        api_client,
        url,
        data=add_forge_data,
        status_code=201,
    )

    date_after = datetime.datetime.now(tz=datetime.timezone.utc)

    consent = add_forge_data["submitter_forward_username"]
    # map the expected result with what's expectedly read from the db to ease comparison
    expected_consent_bool = consent == "on" if isinstance(consent, str) else consent

    assert resp.data == {
        **add_forge_data,
        "id": resp.data["id"],
        "status": "PENDING",
        "submission_date": resp.data["submission_date"],
        "submitter_name": regular_user.username,
        "submitter_email": regular_user.email,
        "submitter_forward_username": expected_consent_bool,
        "last_moderator": resp.data["last_moderator"],
        "last_modified_date": resp.data["last_modified_date"],
        "inbound_email_address": inbound_email_for_pk(resp.data["id"]),
        "forge_domain": urlparse(add_forge_data["forge_url"]).netloc,
    }

    assert date_before < iso8601.parse_date(resp.data["submission_date"]) < date_after

    request = Request.objects.all().last()

    assert request.forge_url == add_forge_data["forge_url"]
    assert request.submitter_name == regular_user.username


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_create_success_form_encoded(client, regular_user):
    client.force_login(regular_user)
    url = reverse("api-1-add-forge-request-create")

    date_before = datetime.datetime.now(tz=datetime.timezone.utc)

    resp = check_http_post_response(
        client,
        url,
        request_content_type="application/x-www-form-urlencoded",
        data=urlencode(ADD_FORGE_DATA_FORGE1),
        status_code=201,
    )

    date_after = datetime.datetime.now(tz=datetime.timezone.utc)

    assert resp.data == {
        **ADD_FORGE_DATA_FORGE1,
        "id": resp.data["id"],
        "status": "PENDING",
        "submission_date": resp.data["submission_date"],
        "submitter_name": regular_user.username,
        "submitter_email": regular_user.email,
        "last_moderator": resp.data["last_moderator"],
        "last_modified_date": resp.data["last_modified_date"],
        "inbound_email_address": inbound_email_for_pk(resp.data["id"]),
        "forge_domain": urlparse(ADD_FORGE_DATA_FORGE1["forge_url"]).netloc,
    }

    assert date_before < iso8601.parse_date(resp.data["submission_date"]) < date_after

    request = Request.objects.all()[0]

    assert request.forge_url == ADD_FORGE_DATA_FORGE1["forge_url"]
    assert request.submitter_name == regular_user.username


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_create_duplicate(api_client, regular_user):
    api_client.force_login(regular_user)
    url = reverse("api-1-add-forge-request-create")
    check_api_post_response(
        api_client,
        url,
        data=ADD_FORGE_DATA_FORGE1,
        status_code=201,
    )
    check_api_post_response(
        api_client,
        url,
        data=ADD_FORGE_DATA_FORGE1,
        status_code=409,
    )

    requests = Request.objects.all()
    assert len(requests) == 1


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_create_invalid_forge_url(api_client, regular_user):
    api_client.force_login(regular_user)
    url = reverse("api-1-add-forge-request-create")

    forge_data = copy.deepcopy(ADD_FORGE_DATA_FORGE1)
    forge_data["forge_url"] = "foo"

    resp = check_api_post_response(
        api_client,
        url,
        data=forge_data,
        status_code=400,
    )

    assert resp.data == {
        "exception": "BadInputExc",
        "reason": '{"forge_url": ["Enter a valid URL."]}',
    }


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_update_anonymous_user(api_client):
    url = reverse("api-1-add-forge-request-update", url_args={"id": 1})
    check_api_post_response(api_client, url, status_code=403)


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_update_regular_user(api_client, regular_user):
    api_client.force_login(regular_user)
    url = reverse("api-1-add-forge-request-update", url_args={"id": 1})
    check_api_post_response(api_client, url, status_code=403)


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_update_non_existent(api_client, add_forge_moderator):
    api_client.force_login(add_forge_moderator)
    url = reverse("api-1-add-forge-request-update", url_args={"id": 1})
    check_api_post_response(api_client, url, status_code=400)


def create_add_forge_request(
    api_client, regular_user, data=ADD_FORGE_DATA_FORGE1, **kwargs
):
    api_client.force_login(regular_user)
    url = reverse("api-1-add-forge-request-create")
    return check_api_post_response(
        api_client, url, data=data, status_code=201, **kwargs
    )


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_update_empty(api_client, regular_user, add_forge_moderator):
    create_add_forge_request(api_client, regular_user)

    api_client.force_login(add_forge_moderator)
    url = reverse("api-1-add-forge-request-update", url_args={"id": 1})
    check_api_post_response(api_client, url, status_code=400)


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_update_missing_field(
    api_client, regular_user, add_forge_moderator
):
    create_add_forge_request(api_client, regular_user)

    api_client.force_login(add_forge_moderator)
    url = reverse("api-1-add-forge-request-update", url_args={"id": 1})
    check_api_post_response(api_client, url, data={}, status_code=400)
    check_api_post_response(
        api_client, url, data={"new_status": "REJECTED"}, status_code=400
    )


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_update(api_client, regular_user, add_forge_moderator):
    resp = create_add_forge_request(api_client, regular_user)

    api_client.force_login(add_forge_moderator)
    url = reverse("api-1-add-forge-request-update", url_args={"id": resp.data["id"]})

    check_api_post_response(
        api_client, url, data={"text": "updating request"}, status_code=200
    )

    check_api_post_response(
        api_client,
        url,
        data={"new_status": "REJECTED", "text": "request rejected"},
        status_code=200,
    )


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_update_invalid_new_status(
    api_client, regular_user, add_forge_moderator
):
    resp = create_add_forge_request(api_client, regular_user)

    api_client.force_login(add_forge_moderator)
    url = reverse("api-1-add-forge-request-update", url_args={"id": resp.data["id"]})
    check_api_post_response(
        api_client,
        url,
        data={"new_status": "ACCEPTED", "text": "request accepted"},
        status_code=400,
    )


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_update_status_concurrent(
    api_client, regular_user, add_forge_moderator, mocker
):
    _block_while_testing = mocker.patch(
        "swh.web.add_forge_now.api_views._block_while_testing"
    )
    _block_while_testing.side_effect = lambda: time.sleep(1)

    resp = create_add_forge_request(api_client, regular_user)

    api_client.force_login(add_forge_moderator)
    url = reverse("api-1-add-forge-request-update", url_args={"id": resp.data["id"]})

    worker_ended = False

    def worker():
        nonlocal worker_ended
        check_api_post_response(
            api_client,
            url,
            data={"new_status": "WAITING_FOR_FEEDBACK", "text": "waiting for message"},
            status_code=200,
        )
        worker_ended = True

    # this thread will first modify the request status to WAITING_FOR_FEEDBACK
    thread = threading.Thread(target=worker)
    thread.start()

    # the other thread (slower) will attempt to modify the request status to REJECTED
    # but it will not be allowed as the first faster thread already modified it
    # and REJECTED state cannot be reached from WAITING_FOR_FEEDBACK one
    time.sleep(0.5)
    check_api_post_response(
        api_client,
        url,
        data={"new_status": "PENDING", "text": "request accepted"},
        status_code=400,
    )
    thread.join()
    assert worker_ended


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_list_anonymous(api_client, regular_user):
    url = reverse("api-1-add-forge-request-list")

    resp = check_api_get_responses(api_client, url, status_code=200)

    assert resp.data == []

    create_add_forge_request(api_client, regular_user)

    resp = check_api_get_responses(api_client, url, status_code=200)

    add_forge_request = {
        "forge_url": ADD_FORGE_DATA_FORGE1["forge_url"],
        "forge_type": ADD_FORGE_DATA_FORGE1["forge_type"],
        "status": "PENDING",
        "submission_date": resp.data[0]["submission_date"],
        "id": resp.data[0]["id"],
    }

    assert resp.data == [add_forge_request]

    create_add_forge_request(api_client, regular_user, data=ADD_FORGE_DATA_FORGE2)

    resp = check_api_get_responses(api_client, url, status_code=200)

    other_forge_request = {
        "forge_url": ADD_FORGE_DATA_FORGE2["forge_url"],
        "forge_type": ADD_FORGE_DATA_FORGE2["forge_type"],
        "status": "PENDING",
        "submission_date": resp.data[0]["submission_date"],
        "id": resp.data[0]["id"],
    }

    assert resp.data == [other_forge_request, add_forge_request]


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_list_moderator(
    api_client, regular_user, add_forge_moderator
):
    url = reverse("api-1-add-forge-request-list")

    create_add_forge_request(api_client, regular_user)
    create_add_forge_request(api_client, regular_user, data=ADD_FORGE_DATA_FORGE2)

    api_client.force_login(add_forge_moderator)
    resp = check_api_get_responses(api_client, url, status_code=200)

    add_forge_request = {
        **ADD_FORGE_DATA_FORGE1,
        "status": "PENDING",
        "submission_date": resp.data[1]["submission_date"],
        "submitter_name": regular_user.username,
        "submitter_email": regular_user.email,
        "last_moderator": resp.data[1]["last_moderator"],
        "last_modified_date": resp.data[1]["last_modified_date"],
        "id": resp.data[1]["id"],
        "inbound_email_address": inbound_email_for_pk(resp.data[1]["id"]),
        "forge_domain": urlparse(ADD_FORGE_DATA_FORGE1["forge_url"]).netloc,
    }

    other_forge_request = {
        **ADD_FORGE_DATA_FORGE2,
        "status": "PENDING",
        "submission_date": resp.data[0]["submission_date"],
        "submitter_name": regular_user.username,
        "submitter_email": regular_user.email,
        "last_moderator": resp.data[0]["last_moderator"],
        "last_modified_date": resp.data[0]["last_modified_date"],
        "id": resp.data[0]["id"],
        "inbound_email_address": inbound_email_for_pk(resp.data[0]["id"]),
        "forge_domain": urlparse(ADD_FORGE_DATA_FORGE2["forge_url"]).netloc,
    }

    assert resp.data == [other_forge_request, add_forge_request]


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_list_pagination(
    api_client, regular_user, api_request_factory
):
    create_add_forge_request(api_client, regular_user)
    create_add_forge_request(api_client, regular_user, data=ADD_FORGE_DATA_FORGE2)

    url = reverse("api-1-add-forge-request-list", query_params={"per_page": 1})

    resp = check_api_get_responses(api_client, url, 200)

    assert len(resp.data) == 1

    request = api_request_factory.get(url)

    next_url = reverse(
        "api-1-add-forge-request-list",
        query_params={"page": 2, "per_page": 1},
        request=request,
    )

    assert resp["Link"] == f'<{next_url}>; rel="next"'

    resp = check_api_get_responses(api_client, next_url, 200)

    assert len(resp.data) == 1

    prev_url = reverse(
        "api-1-add-forge-request-list",
        query_params={"page": 1, "per_page": 1},
        request=request,
    )

    assert resp["Link"] == f'<{prev_url}>; rel="previous"'


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_list_submitter_filtering(
    api_client, regular_user, regular_user2
):
    create_add_forge_request(api_client, regular_user)
    create_add_forge_request(api_client, regular_user2, data=ADD_FORGE_DATA_FORGE2)

    api_client.force_login(regular_user)
    url = reverse(
        "api-1-add-forge-request-list", query_params={"user_requests_only": 1}
    )
    resp = check_api_get_responses(api_client, url, status_code=200)

    assert len(resp.data) == 1


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_get(api_client, regular_user, add_forge_moderator):
    request = create_add_forge_request(api_client, regular_user).data

    submission_date = request["submission_date"]

    url = reverse("api-1-add-forge-request-update", url_args={"id": request["id"]})

    api_client.force_login(add_forge_moderator)
    resp = check_api_post_response(
        api_client,
        url,
        data={"new_status": "WAITING_FOR_FEEDBACK", "text": "waiting for message"},
        status_code=200,
    )
    api_client.logout()

    url = reverse("api-1-add-forge-request-get", url_args={"id": request["id"]})

    resp = check_api_get_responses(api_client, url, status_code=200)

    assert resp.data == {
        "request": {
            "forge_url": ADD_FORGE_DATA_FORGE1["forge_url"],
            "forge_type": ADD_FORGE_DATA_FORGE1["forge_type"],
            "id": request["id"],
            "status": "WAITING_FOR_FEEDBACK",
            "submission_date": submission_date,
        },
        "history": [
            {
                "id": RequestHistory.objects.first().id,
                "actor_role": "SUBMITTER",
                "date": resp.data["history"][0]["date"],
                "new_status": "PENDING",
            },
            {
                "id": RequestHistory.objects.last().id,
                "actor_role": "MODERATOR",
                "date": resp.data["history"][1]["date"],
                "new_status": "WAITING_FOR_FEEDBACK",
            },
        ],
    }


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_get_moderator(api_client, regular_user, add_forge_moderator):
    request = create_add_forge_request(api_client, regular_user).data

    submission_date = request["submission_date"]

    url = reverse("api-1-add-forge-request-update", url_args={"id": request["id"]})

    api_client.force_login(add_forge_moderator)
    check_api_post_response(
        api_client,
        url,
        data={"new_status": "WAITING_FOR_FEEDBACK", "text": "waiting for message"},
        status_code=200,
    )

    url = reverse("api-1-add-forge-request-get", url_args={"id": request["id"]})

    resp = check_api_get_responses(api_client, url, status_code=200)
    resp.data["history"] = [dict(history_item) for history_item in resp.data["history"]]

    assert resp.data == {
        "request": {
            **ADD_FORGE_DATA_FORGE1,
            "id": request["id"],
            "status": "WAITING_FOR_FEEDBACK",
            "submission_date": submission_date,
            "submitter_name": regular_user.username,
            "submitter_email": regular_user.email,
            "last_moderator": add_forge_moderator.username,
            "last_modified_date": resp.data["history"][1]["date"],
            "inbound_email_address": inbound_email_for_pk(request["id"]),
            "forge_domain": urlparse(ADD_FORGE_DATA_FORGE1["forge_url"]).netloc,
        },
        "history": [
            {
                "id": RequestHistory.objects.first().id,
                "text": "",
                "actor": regular_user.username,
                "actor_role": "SUBMITTER",
                "date": resp.data["history"][0]["date"],
                "new_status": "PENDING",
                "message_source_url": None,
            },
            {
                "id": RequestHistory.objects.last().id,
                "text": "waiting for message",
                "actor": add_forge_moderator.username,
                "actor_role": "MODERATOR",
                "date": resp.data["history"][1]["date"],
                "new_status": "WAITING_FOR_FEEDBACK",
                "message_source_url": None,
            },
        ],
    }


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_get_moderator_message_source(
    api_client, regular_user, add_forge_moderator
):
    resp = create_add_forge_request(api_client, regular_user)

    rh = RequestHistory(
        request=Request.objects.get(pk=resp.data["id"]),
        new_status="WAITING_FOR_FEEDBACK",
        text="waiting for message",
        actor=add_forge_moderator.username,
        actor_role="MODERATOR",
        message_source=b"test with a message source",
    )
    rh.save()

    api_client.force_login(add_forge_moderator)
    url = reverse("api-1-add-forge-request-get", url_args={"id": resp.data["id"]})
    resp = check_api_get_responses(api_client, url, status_code=200)
    resp.data["history"] = [dict(history_item) for history_item in resp.data["history"]]

    # Check that the authentified moderator can't urlhack non-existent message sources
    assert resp.data["history"][0]["message_source_url"] is None
    empty_message_url = reverse(
        "forge-add-message-source", url_args={"id": resp.data["history"][0]["id"]}
    )
    check_http_get_response(api_client, empty_message_url, status_code=404)

    # Check that the authentified moderator can't urlhack non-existent message sources
    non_existent_message_url = reverse(
        "forge-add-message-source", url_args={"id": 9001}
    )
    check_http_get_response(api_client, non_existent_message_url, status_code=404)

    # Check that the authentified moderator can access the message source when the url is
    # given

    message_source_url = resp.data["history"][-1]["message_source_url"]
    assert message_source_url is not None

    message_source_resp = check_http_get_response(
        api_client, message_source_url, status_code=200, content_type="text/email"
    )

    # Check that the message source shows up as an attachment
    assert message_source_resp.content == rh.message_source
    disposition = message_source_resp["Content-Disposition"]
    assert disposition.startswith("attachment; filename=")
    assert disposition.endswith('.eml"')

    # Check that a regular user can't access message sources
    api_client.force_login(regular_user)
    check_http_get_response(api_client, message_source_url, status_code=302)

    api_client.force_login(add_forge_moderator)


@pytest.mark.django_db(transaction=True)
def test_add_forge_request_get_invalid(api_client):
    url = reverse("api-1-add-forge-request-get", url_args={"id": 3})
    check_api_get_responses(api_client, url, status_code=400)


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize("success", [True, False], ids=["success", "failure"])
def test_add_forge_request_update_gitlab_pipeline_trigger(
    api_client, regular_user, add_forge_moderator, requests_mock, mocker, success
):
    trigger_url = (
        "https://gitlab.softwareheritage.org/api/v4/projects/xyz/trigger/pipeline"
    )
    token = "some-token"

    # patch configuration to enable the POST request triggering gitlab pipeline
    config = copy.deepcopy(get_config())
    config["add_forge_now"]["gitlab_pipeline"] = {
        "trigger_url": trigger_url,
        "token": token,
    }
    mock_get_config = mocker.patch("swh.web.add_forge_now.utils.get_config")
    mock_get_config.return_value = config

    # mock POST request sent to gitlab
    if success:
        requests_mock.post(trigger_url, text="OK")
    else:
        requests_mock.post(trigger_url, text="KO", status_code=403)

    request = create_add_forge_request(api_client, regular_user).data

    api_client.force_login(add_forge_moderator)
    url = reverse("api-1-add-forge-request-update", url_args={"id": request["id"]})

    check_api_post_response(
        api_client,
        url,
        data={"new_status": "WAITING_FOR_FEEDBACK", "text": "waiting for feedback"},
        status_code=200,
    )

    resp = check_api_post_response(
        api_client,
        url,
        data={"new_status": "ACCEPTED", "text": "request accepted"},
        status_code=200 if success else 500,
    )

    # check POST request was sent
    assert requests_mock.request_history
    sent_request = requests_mock.request_history[0]
    assert sent_request.url == trigger_url
    assert sent_request.headers["Content-Type"] == "application/x-www-form-urlencoded"

    id = request["id"]
    lister_type = ADD_FORGE_DATA_FORGE1["forge_type"]
    forge_url = ADD_FORGE_DATA_FORGE1["forge_url"]
    instance_name = urlparse(ADD_FORGE_DATA_FORGE1["forge_url"]).netloc
    expected_form_data = {
        "token": token,
        "ref": "main",
        "name": f"AFNR#{id} {lister_type} {instance_name}",
        "variables[LISTER_TYPE]": lister_type,
        "variables[FORGE_URL]": forge_url,
        "variables[INSTANCE_NAME]": instance_name,
        "variables[REQUEST_ID]": id,
    }

    assert requests_mock.request_history[0].body == urlencode(expected_form_data)

    if not success:
        # check POST request error is reported
        assert resp.data == {
            "exception": "HTTPError",
            "reason": f"403 Client Error: None for url: {trigger_url}",
        }



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


def test_create_users_test_users_exist(db):
    from .create_test_users import User, users

    for username, (_, _, permissions, groups) in users.items():
        user = User.objects.filter(username=username).get()
        assert user is not None

        for permission in permissions:
            assert user.has_perm(permission)

        for group in groups:
            assert user.groups.filter(name=group).exists()



# Copyright (C) 2017-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import re

import pytest

from swh.model.model import Content
from swh.web.browse.utils import (
    gen_link,
    gen_person_mail_link,
    gen_revision_link,
    get_mimetype_and_encoding_for_content,
    get_readme_to_display,
    prepare_content_for_display,
    re_encode_content,
)
from swh.web.tests.data import get_tests_data
from swh.web.utils import reverse


def test_get_mimetype_and_encoding_for_content():
    text = b"Hello world!"
    assert get_mimetype_and_encoding_for_content(text) == (
        "text/plain",
        "us-ascii",
    )


def test_gen_link():
    assert (
        gen_link("https://www.softwareheritage.org/", "swh")
        == '<a href="https://www.softwareheritage.org/">swh</a>'
    )


def test_gen_revision_link():
    revision_id = "28a0bc4120d38a394499382ba21d6965a67a3703"
    revision_url = reverse("browse-revision", url_args={"sha1_git": revision_id})

    assert gen_revision_link(
        revision_id, link_text=None, link_attrs=None
    ) == '<a href="%s">%s</a>' % (revision_url, revision_id)
    assert gen_revision_link(
        revision_id, shorten_id=True, link_attrs=None
    ) == '<a href="%s">%s</a>' % (revision_url, revision_id[:7])


def test_gen_person_mail_link():
    person_full = {
        "name": "John Doe",
        "email": "john.doe@swh.org",
        "fullname": "John Doe <john.doe@swh.org>",
    }

    assert gen_person_mail_link(person_full) == '<a href="mailto:%s">%s</a>' % (
        person_full["email"],
        person_full["name"],
    )

    link_text = "Mail"
    assert gen_person_mail_link(
        person_full, link_text=link_text
    ) == '<a href="mailto:%s">%s</a>' % (person_full["email"], link_text)

    person_partial_email = {"name": None, "email": None, "fullname": "john.doe@swh.org"}

    assert gen_person_mail_link(
        person_partial_email
    ) == '<a href="mailto:%s">%s</a>' % (
        person_partial_email["fullname"],
        person_partial_email["fullname"],
    )

    person_partial = {
        "name": None,
        "email": None,
        "fullname": "John Doe <john.doe@swh.org>",
    }

    assert gen_person_mail_link(person_partial) == person_partial["fullname"]

    person_none = {"name": None, "email": None, "fullname": None}

    assert gen_person_mail_link(person_none) == "None"


@pytest.mark.parametrize(
    "path, expected_language",
    [("CMakeLists.txt", "cmake"), ("path/CMakeLists.txt", "cmake")],
)
def test_prepare_content_display_language_for_filename(path, expected_language):
    content_display = prepare_content_for_display(
        content_data=b"", mime_type="", encoding="", path=path
    )
    assert content_display["language"] == expected_language


def test_re_encode_content_for_shift_jis_encoding():
    data = b"/* \x8a\xd6\x98A\x82\xcc\x95\xb6\x8e\x9a\x83R\x81[\x83h\x95\xcf\x8a\xb7 */"
    mime_type, encoding = get_mimetype_and_encoding_for_content(data)

    _, encoding, re_encoded_data = re_encode_content(mime_type, encoding, data)

    assert encoding == "SHIFT_JIS"
    assert data.decode(encoding) == re_encoded_data.decode("utf-8")
    assert re_encoded_data.decode("utf-8") == "/* 関連の文字コード変換 */"


def test_re_encode_content_chardet_fallback(datadir):
    with open(os.path.join(datadir, "content_windows-1253_encoded"), "rb") as f:
        content = f.read()
        mime_type, encoding = get_mimetype_and_encoding_for_content(content)
        _, encoding, _ = re_encode_content(mime_type, encoding, content)
        assert encoding == "WINDOWS-1253"


@pytest.mark.parametrize(
    "input_,expected_output",
    [
        (b"foo bar", "<p>foo bar</p>"),
        (b"foo *bar* baz", "<p>foo <em>bar</em> baz</p>"),
        (
            b".. raw:: html\n\n   <script>foo</script>",
            "&lt;script&gt;foo&lt;/script&gt;",
        ),
    ],
)
def test_rst_readme(input_, expected_output):
    content = Content.from_data(input_)
    storage = get_tests_data()["storage"]
    storage.content_add([content])
    assert re.search(
        expected_output,
        get_readme_to_display({"readme.rst": content.sha1_git.hex()})[2],
    )


def test_rst_readme_no_leak():
    input_ = b".. include:: /etc/passwd"
    content = Content.from_data(input_)
    storage = get_tests_data()["storage"]
    storage.content_add([content])
    assert (
        "root:" not in get_readme_to_display({"readme.rst": content.sha1_git.hex()})[2]
    )


def test_rst_readme_no_xss():
    input_ = b".. raw:: html\n\n   <script>foo</script>"
    content = Content.from_data(input_)
    storage = get_tests_data()["storage"]
    storage.content_add([content])
    assert (
        "<script>"
        not in get_readme_to_display({"readme.rst": content.sha1_git.hex()})[2]
    )



# Copyright (C) 2017-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import timedelta
import random
import re

from hypothesis import given
import pytest

from django.test.utils import override_settings
from django.utils.html import escape

from swh.model.from_disk import DentryPerms
from swh.model.hashutil import hash_to_bytes
from swh.model.model import (
    Content,
    Directory,
    DirectoryEntry,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    Release,
    ReleaseTargetType,
    Snapshot,
    SnapshotBranch,
    SnapshotTargetType,
)
from swh.model.swhids import ObjectType
from swh.storage.utils import now
from swh.web.browse.snapshot_context import process_snapshot_branches
from swh.web.browse.utils import (
    get_mimetype_and_encoding_for_content,
    prepare_content_for_display,
    re_encode_content,
)
from swh.web.tests.data import get_content
from swh.web.tests.django_asserts import assert_contains, assert_not_contains
from swh.web.tests.helpers import check_html_get_response, check_http_get_response
from swh.web.tests.strategies import new_origin
from swh.web.utils import (
    format_utc_iso_date,
    gen_path_info,
    parse_iso8601_date_to_utc,
    reverse,
)
from swh.web.utils.exc import NotFoundExc
from swh.web.utils.identifiers import gen_swhid


def test_content_view_text(client, archive_data, content_text):
    sha1_git = content_text["sha1_git"]

    url = reverse(
        "browse-content",
        url_args={"query_string": content_text["sha1"]},
        query_params={"path": content_text["path"]},
    )

    url_raw = reverse(
        "browse-content-raw", url_args={"query_string": content_text["sha1"]}
    )

    resp = check_html_get_response(
        client, url, status_code=200, template_used="browse-content.html"
    )

    content_display = _process_content_for_display(archive_data, content_text)
    mimetype = content_display["mimetype"]

    if mimetype.startswith("text/"):
        assert_contains(resp, '<code class="%s">' % content_display["language"])
        assert_contains(resp, escape(content_display["content_data"]))
    assert_contains(resp, url_raw)

    swh_cnt_id = gen_swhid(ObjectType.CONTENT, sha1_git)
    swh_cnt_id_url = reverse("browse-swhid", url_args={"swhid": swh_cnt_id})
    assert_contains(resp, swh_cnt_id)
    assert_contains(resp, swh_cnt_id_url)
    assert_not_contains(resp, "swh-metadata-popover")


def test_content_view_no_highlight(
    client, archive_data, content_application_no_highlight, content_text_no_highlight
):
    for content_ in (content_application_no_highlight, content_text_no_highlight):
        content = content_
        sha1_git = content["sha1_git"]

        url = reverse("browse-content", url_args={"query_string": content["sha1"]})

        url_raw = reverse(
            "browse-content-raw", url_args={"query_string": content["sha1"]}
        )

        resp = check_html_get_response(
            client, url, status_code=200, template_used="browse-content.html"
        )

        content_display = _process_content_for_display(archive_data, content)

        if content["encoding"] != "binary":
            assert_contains(resp, '<code class="plaintext">')
            assert_contains(resp, escape(content_display["content_data"]))

        assert_contains(resp, url_raw)

        swh_cnt_id = gen_swhid(ObjectType.CONTENT, sha1_git)
        swh_cnt_id_url = reverse("browse-swhid", url_args={"swhid": swh_cnt_id})

        assert_contains(resp, swh_cnt_id)
        assert_contains(resp, swh_cnt_id_url)


def test_content_view_no_utf8_text(client, archive_data, content_text_non_utf8):
    sha1_git = content_text_non_utf8["sha1_git"]

    url = reverse(
        "browse-content", url_args={"query_string": content_text_non_utf8["sha1"]}
    )

    resp = check_html_get_response(
        client, url, status_code=200, template_used="browse-content.html"
    )

    content_display = _process_content_for_display(archive_data, content_text_non_utf8)

    swh_cnt_id = gen_swhid(ObjectType.CONTENT, sha1_git)
    swh_cnt_id_url = reverse("browse-swhid", url_args={"swhid": swh_cnt_id})
    assert_contains(resp, swh_cnt_id_url)
    assert_contains(resp, escape(content_display["content_data"]))


def test_content_view_image(client, archive_data, content_image_type):
    url = reverse(
        "browse-content", url_args={"query_string": content_image_type["sha1"]}
    )

    url_raw = reverse(
        "browse-content-raw", url_args={"query_string": content_image_type["sha1"]}
    )

    resp = check_html_get_response(
        client, url, status_code=200, template_used="browse-content.html"
    )

    content_display = _process_content_for_display(archive_data, content_image_type)
    mimetype = content_display["mimetype"]
    content_data = content_display["content_data"]
    assert_contains(
        resp, f'<img src="data:{mimetype};base64,{content_data}" alt="image" />'
    )
    assert_contains(resp, url_raw)


def test_content_view_image_no_rendering(
    client, archive_data, content_unsupported_image_type_rendering
):
    url = reverse(
        "browse-content",
        url_args={"query_string": content_unsupported_image_type_rendering["sha1"]},
    )

    resp = check_html_get_response(
        client, url, status_code=200, template_used="browse-content.html"
    )

    mimetype = content_unsupported_image_type_rendering["mimetype"]
    encoding = content_unsupported_image_type_rendering["encoding"]
    assert_contains(
        resp,
        (
            f"Content with mime type {mimetype} and encoding {encoding} "
            "cannot be displayed."
        ),
    )


def test_content_view_text_with_path(client, archive_data, content_text):
    path = content_text["path"]

    url = reverse(
        "browse-content",
        url_args={"query_string": content_text["sha1"]},
        query_params={"path": path},
    )

    resp = check_html_get_response(
        client, url, status_code=200, template_used="browse-content.html"
    )

    assert_contains(resp, '<nav class="bread-crumbs')

    content_display = _process_content_for_display(archive_data, content_text)
    mimetype = content_display["mimetype"]

    if mimetype.startswith("text/"):
        hljs_language = content_text["hljs_language"]
        assert_contains(resp, '<code class="%s">' % hljs_language)
        assert_contains(resp, escape(content_display["content_data"]))

    split_path = path.split("/")

    root_dir_sha1 = split_path[0]
    filename = split_path[-1]
    path = path.replace(root_dir_sha1 + "/", "").replace(filename, "")

    swhid_context = {
        "anchor": gen_swhid(ObjectType.DIRECTORY, root_dir_sha1),
        "path": f"/{path}{filename}",
    }

    swh_cnt_id = gen_swhid(
        ObjectType.CONTENT, content_text["sha1_git"], metadata=swhid_context
    )
    swh_cnt_id_url = reverse("browse-swhid", url_args={"swhid": swh_cnt_id})
    assert_contains(resp, swh_cnt_id)
    assert_contains(resp, swh_cnt_id_url)

    path_info = gen_path_info(path)

    root_dir_url = reverse("browse-directory", url_args={"sha1_git": root_dir_sha1})

    assert_contains(resp, '<li class="swh-path">', count=len(path_info) + 1)

    assert_contains(
        resp, '<a href="' + root_dir_url + '">' + root_dir_sha1[:7] + "</a>"
    )

    for p in path_info:
        dir_url = reverse(
            "browse-directory",
            url_args={"sha1_git": root_dir_sha1},
            query_params={"path": p["path"]},
        )
        assert_contains(resp, '<a href="' + dir_url + '">' + p["name"] + "</a>")

    assert_contains(resp, "<li>" + filename + "</li>")

    url_raw = reverse(
        "browse-content-raw",
        url_args={"query_string": content_text["sha1"]},
        query_params={"filename": filename},
    )
    assert_contains(resp, url_raw)

    url = reverse(
        "browse-content",
        url_args={"query_string": content_text["sha1"]},
        query_params={"path": filename},
    )

    resp = check_html_get_response(
        client, url, status_code=200, template_used="browse-content.html"
    )

    assert_not_contains(resp, '<nav class="bread-crumbs')

    invalid_path = "%s/foo/bar/baz" % root_dir_sha1
    url = reverse(
        "browse-content",
        url_args={"query_string": content_text["sha1"]},
        query_params={"path": invalid_path},
    )

    resp = check_html_get_response(
        client, url, status_code=404, template_used="error.html"
    )


@pytest.mark.parametrize("filename_prefix", ["", "/home/"])
def test_content_raw_text(client, archive_data, content_text, filename_prefix):
    url = reverse("browse-content-raw", url_args={"query_string": content_text["sha1"]})

    resp = check_http_get_response(
        client, url, status_code=200, content_type="text/plain"
    )

    content_data = archive_data.content_get_data(content_text["sha1"])["data"]

    assert resp["Content-Type"] == "text/plain"
    assert resp["Content-disposition"] == (f'filename="sha1_{content_text["sha1"]}"')
    assert b"".join(resp.streaming_content) == content_data

    filename = content_text["path"].split("/")[-1]

    url = reverse(
        "browse-content-raw",
        url_args={"query_string": content_text["sha1"]},
        query_params={"filename": filename_prefix + filename},
    )

    resp = check_http_get_response(
        client, url, status_code=200, content_type="text/plain"
    )

    assert resp["Content-Type"] == "text/plain"
    assert resp["Content-disposition"] == f'filename="{filename}"'
    assert b"".join(resp.streaming_content) == content_data


def test_content_raw_no_utf8_text(client, content_text_non_utf8):
    url = reverse(
        "browse-content-raw", url_args={"query_string": content_text_non_utf8["sha1"]}
    )

    resp = check_http_get_response(
        client, url, status_code=200, content_type="text/plain"
    )
    _, encoding = get_mimetype_and_encoding_for_content(
        b"".join(resp.streaming_content)
    )
    assert encoding == content_text_non_utf8["encoding"]


def test_content_raw_bin(client, archive_data, content_binary_type):
    url = reverse(
        "browse-content-raw", url_args={"query_string": content_binary_type["sha1"]}
    )

    resp = check_http_get_response(
        client, url, status_code=200, content_type="application/octet-stream"
    )

    filename = content_binary_type["path"].split("/")[-1]
    content_data = archive_data.content_get_data(content_binary_type["sha1"])["data"]

    assert resp["Content-Type"] == "application/octet-stream"
    assert (
        resp["Content-disposition"]
        == f'attachment; filename="sha1_{content_binary_type["sha1"]}"'
    )
    assert b"".join(resp.streaming_content) == content_data

    url = reverse(
        "browse-content-raw",
        url_args={"query_string": content_binary_type["sha1"]},
        query_params={"filename": filename},
    )

    resp = check_http_get_response(
        client, url, status_code=200, content_type="application/octet-stream"
    )

    assert resp["Content-Type"] == "application/octet-stream"
    assert resp["Content-disposition"] == f'attachment; filename="{filename}"'
    assert b"".join(resp.streaming_content) == content_data


def test_content_raw_image_with_content_type(client, archive_data, content_image_type):
    url_raw = reverse(
        "browse-content-raw",
        url_args={"query_string": content_image_type["sha1"]},
    )

    content_display = _process_content_for_display(archive_data, content_image_type)

    assert content_display["mimetype"].startswith("image/")

    check_http_get_response(
        client, url_raw, status_code=200, content_type=content_display["mimetype"]
    )


def test_content_raw_re_encode_invalid_param(client, content_text):
    url = reverse(
        "browse-content-raw",
        url_args={"query_string": content_text["sha1"]},
        query_params={"re_encode": "true'"},
    )

    check_http_get_response(
        client,
        url,
        status_code=400,
    )


@pytest.mark.django_db
@pytest.mark.parametrize("staff_user_logged_in", [False, True])
def test_content_request_errors(
    client, staff_user, invalid_sha1, unknown_content, staff_user_logged_in
):
    if staff_user_logged_in:
        client.force_login(staff_user)

    url = reverse("browse-content", url_args={"query_string": invalid_sha1})
    check_html_get_response(client, url, status_code=400, template_used="error.html")

    url = reverse("browse-content", url_args={"query_string": unknown_content["sha1"]})
    check_html_get_response(
        client, url, status_code=404, template_used="browse-content.html"
    )


def test_content_bytes_missing(client, archive_data, mocker, content):
    mock_archive = mocker.patch("swh.web.browse.utils.archive")
    content_data = archive_data.content_get(content["sha1"])

    mock_archive.lookup_content.return_value = content_data
    mock_archive.lookup_content_filetype.side_effect = Exception()
    mock_archive.lookup_content.side_effect = NotFoundExc(
        "Content bytes not available!"
    )

    url = reverse("browse-content", url_args={"query_string": content["sha1"]})

    check_html_get_response(
        client, url, status_code=404, template_used="browse-content.html"
    )


def test_content_masked(
    client, archive_data, mocker, content, make_masked_object_exception
):
    masked_object_exception = make_masked_object_exception(
        f"swh:1:cnt:{content['sha1_git']}"
    )

    mock_archive = mocker.patch("swh.web.browse.utils.archive")
    mock_archive.lookup_content.side_effect = masked_object_exception

    url = reverse("browse-content", url_args={"query_string": content["sha1"]})

    check_html_get_response(client, url, status_code=403, template_used="masked.html")
    check_http_get_response(
        client, url, content_type="application/json", status_code=403
    )
    check_http_get_response(
        client, url, content_type="application/yaml", status_code=403
    )


def test_content_too_large(client, mocker):
    mock_request_content = mocker.patch("swh.web.browse.views.content.request_content")
    stub_content_too_large_data = {
        "checksums": {
            "sha1": "8624bcdae55baeef00cd11d5dfcfa60f68710a02",
            "sha1_git": "94a9ed024d3859793618152ea559a168bbcbb5e2",
            "sha256": (
                "8ceb4b9ee5adedde47b31e975c1d90c73ad27b6b16" "5a1dcd80c7c545eb65b903"
            ),
            "blake2s256": (
                "38702b7168c7785bfe748b51b45d9856070ba90" "f9dc6d90f2ea75d4356411ffe"
            ),
        },
        "length": 30000000,
        "raw_data": None,
        "mimetype": "text/plain",
        "encoding": "us-ascii",
        "language": "not detected",
        "licenses": "GPL",
        "error_code": 200,
        "error_message": "",
        "error_description": "",
    }

    content_sha1 = stub_content_too_large_data["checksums"]["sha1"]

    mock_request_content.return_value = stub_content_too_large_data

    url = reverse("browse-content", url_args={"query_string": content_sha1})

    url_raw = reverse("browse-content-raw", url_args={"query_string": content_sha1})

    resp = check_html_get_response(
        client, url, status_code=200, template_used="browse-content.html"
    )

    assert_contains(resp, "Content is too large to be displayed")
    assert_contains(resp, url_raw)


def test_content_uppercase(client, content):
    url = reverse(
        "browse-content-uppercase-checksum",
        url_args={"query_string": content["sha1"].upper()},
    )

    resp = check_html_get_response(client, url, status_code=302)

    redirect_url = reverse("browse-content", url_args={"query_string": content["sha1"]})

    assert resp["location"] == redirect_url


def test_content_utf8_detected_as_binary_display(
    client, archive_data, content_utf8_detected_as_binary
):
    url = reverse(
        "browse-content",
        url_args={"query_string": content_utf8_detected_as_binary["sha1"]},
    )

    resp = check_html_get_response(
        client, url, status_code=200, template_used="browse-content.html"
    )

    content_display = _process_content_for_display(
        archive_data, content_utf8_detected_as_binary
    )

    assert_contains(resp, escape(content_display["content_data"]))


def test_content_origin_snapshot_branch_browse(
    client, archive_data, origin_with_multiple_visits
):
    origin_url = origin_with_multiple_visits["url"]
    visits = archive_data.origin_visit_get(origin_url)
    visit = random.choice(visits)
    snapshot = archive_data.snapshot_get(visit["snapshot"])
    snapshot_sizes = archive_data.snapshot_count_branches(visit["snapshot"])
    branches, releases, _ = process_snapshot_branches(snapshot)
    branch_info = random.choice(branches)

    directory = archive_data.revision_get(branch_info["target"])["directory"]
    directory_content = archive_data.directory_ls(directory)
    directory_file = random.choice(
        [e for e in directory_content if e["type"] == "file"]
    )

    url = reverse(
        "browse-content",
        url_args={"query_string": directory_file["checksums"]["sha1"]},
        query_params={
            "origin_url": origin_with_multiple_visits["url"],
            "snapshot": snapshot["id"],
            "branch": branch_info["name"],
            "path": directory_file["name"],
        },
    )

    resp = check_html_get_response(
        client, url, status_code=200, template_used="browse-content.html"
    )

    _check_origin_snapshot_related_html(
        resp, origin_with_multiple_visits, snapshot, snapshot_sizes, branches, releases
    )
    assert_contains(resp, directory_file["name"])
    assert_contains(resp, f"Branch: <strong>{branch_info['name']}</strong>")

    cnt_swhid = gen_swhid(
        ObjectType.CONTENT,
        directory_file["checksums"]["sha1_git"],
        metadata={
            "origin": origin_url,
            "visit": gen_swhid(ObjectType.SNAPSHOT, snapshot["id"]),
            "anchor": gen_swhid(ObjectType.REVISION, branch_info["target"]),
            "path": f"/{directory_file['name']}",
        },
    )
    assert_contains(resp, cnt_swhid)

    dir_swhid = gen_swhid(
        ObjectType.DIRECTORY,
        directory,
        metadata={
            "origin": origin_url,
            "visit": gen_swhid(ObjectType.SNAPSHOT, snapshot["id"]),
            "anchor": gen_swhid(ObjectType.REVISION, branch_info["target"]),
        },
    )
    assert_contains(resp, dir_swhid)

    rev_swhid = gen_swhid(
        ObjectType.REVISION,
        branch_info["target"],
        metadata={
            "origin": origin_url,
            "visit": gen_swhid(ObjectType.SNAPSHOT, snapshot["id"]),
        },
    )
    assert_contains(resp, rev_swhid)

    snp_swhid = gen_swhid(
        ObjectType.SNAPSHOT,
        snapshot["id"],
        metadata={
            "origin": origin_url,
        },
    )
    assert_contains(resp, snp_swhid)


def test_content_origin_snapshot_release_browse(
    client, archive_data, origin_with_multiple_visits
):
    origin_url = origin_with_multiple_visits["url"]
    visits = archive_data.origin_visit_get(origin_url)
    visit = random.choice(visits)
    snapshot = archive_data.snapshot_get(visit["snapshot"])
    snapshot_sizes = archive_data.snapshot_count_branches(visit["snapshot"])
    branches, releases, _ = process_snapshot_branches(snapshot)
    release_info = random.choice(releases)

    directory_content = archive_data.directory_ls(release_info["directory"])
    directory_file = random.choice(
        [e for e in directory_content if e["type"] == "file"]
    )

    url = reverse(
        "browse-content",
        url_args={"query_string": directory_file["checksums"]["sha1"]},
        query_params={
            "origin_url": origin_url,
            "snapshot": snapshot["id"],
            "release": release_info["name"],
            "path": directory_file["name"],
        },
    )

    resp = check_html_get_response(
        client, url, status_code=200, template_used="browse-content.html"
    )

    _check_origin_snapshot_related_html(
        resp, origin_with_multiple_visits, snapshot, snapshot_sizes, branches, releases
    )
    assert_contains(resp, directory_file["name"])
    assert_contains(resp, f"Release: <strong>{release_info['name']}</strong>")

    cnt_swhid = gen_swhid(
        ObjectType.CONTENT,
        directory_file["checksums"]["sha1_git"],
        metadata={
            "origin": origin_url,
            "visit": gen_swhid(ObjectType.SNAPSHOT, snapshot["id"]),
            "anchor": gen_swhid(ObjectType.RELEASE, release_info["id"]),
            "path": f"/{directory_file['name']}",
        },
    )
    assert_contains(resp, cnt_swhid)

    dir_swhid = gen_swhid(
        ObjectType.DIRECTORY,
        release_info["directory"],
        metadata={
            "origin": origin_url,
            "visit": gen_swhid(ObjectType.SNAPSHOT, snapshot["id"]),
            "anchor": gen_swhid(ObjectType.RELEASE, release_info["id"]),
        },
    )
    assert_contains(resp, dir_swhid)

    rev_swhid = gen_swhid(
        ObjectType.REVISION,
        release_info["target"],
        metadata={
            "origin": origin_url,
            "visit": gen_swhid(ObjectType.SNAPSHOT, snapshot["id"]),
        },
    )
    assert_contains(resp, rev_swhid)

    rel_swhid = gen_swhid(
        ObjectType.RELEASE,
        release_info["id"],
        metadata={
            "origin": origin_url,
            "visit": gen_swhid(ObjectType.SNAPSHOT, snapshot["id"]),
        },
    )
    assert_contains(resp, rel_swhid)

    snp_swhid = gen_swhid(
        ObjectType.SNAPSHOT,
        snapshot["id"],
        metadata={
            "origin": origin_url,
        },
    )
    assert_contains(resp, snp_swhid)


def _check_origin_snapshot_related_html(
    resp, origin, snapshot, snapshot_sizes, branches, releases
):
    browse_origin_url = reverse(
        "browse-origin", query_params={"origin_url": origin["url"]}
    )
    assert_contains(resp, f'href="{browse_origin_url}"')

    origin_branches_url = reverse(
        "browse-origin-branches",
        query_params={"origin_url": origin["url"], "snapshot": snapshot["id"]},
    )

    assert_contains(resp, f'href="{escape(origin_branches_url)}"')
    assert_contains(resp, f"Branches ({snapshot_sizes['revision']})")

    origin_releases_url = reverse(
        "browse-origin-releases",
        query_params={"origin_url": origin["url"], "snapshot": snapshot["id"]},
    )

    assert_contains(resp, f'href="{escape(origin_releases_url)}"')
    assert_contains(resp, f"Releases ({snapshot_sizes['release']})")

    assert_contains(resp, '<li class="swh-branch"', count=len(branches))
    assert_contains(resp, '<li class="swh-release"', count=len(releases))


def _process_content_for_display(archive_data, content):
    content_data = archive_data.content_get_data(content["sha1"])

    mime_type, encoding = get_mimetype_and_encoding_for_content(content_data["data"])

    mime_type, encoding, content_data = re_encode_content(
        mime_type, encoding, content_data["data"]
    )

    content_display = prepare_content_for_display(
        content_data, mime_type, encoding, content["path"]
    )

    return content_display


def test_content_dispaly_empty_query_string_missing_path(client):
    url = reverse(
        "browse-content",
        query_params={"origin_url": "http://example.com"},
    )
    resp = check_html_get_response(
        client, url, status_code=400, template_used="error.html"
    )
    assert_contains(resp, "The path query parameter must be provided.", status_code=400)


def test_content_dispaly_empty_query_string_and_snapshot_origin(client):
    url = reverse(
        "browse-content",
        query_params={"path": "test.txt"},
    )
    resp = check_html_get_response(
        client,
        url,
        status_code=400,
    )
    assert_contains(
        resp,
        "The origin_url or snapshot query parameters must be provided.",
        status_code=400,
    )


def test_content_dispaly_empty_query_string_with_origin(
    client, archive_data, origin_with_multiple_visits
):
    origin_url = origin_with_multiple_visits["url"]
    snapshot = archive_data.snapshot_get_latest(origin_url)
    head_rev_id = archive_data.snapshot_get_head(snapshot)
    head_rev = archive_data.revision_get(head_rev_id)
    dir_content = archive_data.directory_ls(head_rev["directory"])
    dir_files = [e for e in dir_content if e["type"] == "file"]
    dir_file = random.choice(dir_files)

    url = reverse(
        "browse-content",
        query_params={
            "origin_url": origin_url,
            "path": dir_file["name"],
        },
    )

    resp = check_html_get_response(
        client,
        url,
        status_code=302,
    )
    redict_url = reverse(
        "browse-content",
        url_args={"query_string": f"sha1_git:{dir_file['checksums']['sha1_git']}"},
        query_params={
            "origin_url": origin_url,
            "path": dir_file["name"],
        },
    )
    assert resp.url == redict_url


def test_content_dispaly_empty_query_string_with_snapshot(
    client, archive_data, origin_with_multiple_visits
):
    origin_url = origin_with_multiple_visits["url"]
    snapshot = archive_data.snapshot_get_latest(origin_url)
    head_rev_id = archive_data.snapshot_get_head(snapshot)
    head_rev = archive_data.revision_get(head_rev_id)
    dir_content = archive_data.directory_ls(head_rev["directory"])
    dir_files = [e for e in dir_content if e["type"] == "file"]
    dir_file = random.choice(dir_files)
    url = reverse(
        "browse-content",
        query_params={
            "snapshot": snapshot["id"],
            "path": dir_file["name"],
        },
    )

    resp = check_html_get_response(
        client,
        url,
        status_code=302,
    )
    redict_url = reverse(
        "browse-content",
        url_args={"query_string": f"sha1_git:{dir_file['checksums']['sha1_git']}"},
        query_params={
            "snapshot": snapshot["id"],
            "path": dir_file["name"],
        },
    )
    assert resp.url == redict_url


def test_browse_origin_content_no_visit(client, mocker, origin):
    mock_get_origin_visits = mocker.patch(
        "swh.web.utils.origin_visits.get_origin_visits"
    )
    mock_get_origin_visits.return_value = []
    mock_archive = mocker.patch("swh.web.utils.origin_visits.archive")
    mock_archive.lookup_origin_visit_latest.return_value = None
    url = reverse(
        "browse-content",
        query_params={"origin_url": origin["url"], "path": "foo"},
    )

    resp = check_html_get_response(
        client, url, status_code=404, template_used="error.html"
    )
    assert_contains(resp, "No valid visit", status_code=404)
    assert not mock_get_origin_visits.called


def test_browse_origin_content_unknown_visit(client, mocker, origin):
    mock_get_origin_visits = mocker.patch(
        "swh.web.utils.origin_visits.get_origin_visits"
    )
    mock_get_origin_visits.return_value = [{"visit": 1}]

    url = reverse(
        "browse-content",
        query_params={"origin_url": origin["url"], "path": "foo", "visit_id": 2},
    )

    resp = check_html_get_response(
        client, url, status_code=404, template_used="error.html"
    )
    assert re.search("Resource not found", resp.content.decode("utf-8"))


def test_browse_origin_content_not_found(client, origin):
    url = reverse(
        "browse-content",
        query_params={"origin_url": origin["url"], "path": "/invalid/file/path"},
    )

    resp = check_html_get_response(
        client, url, status_code=404, template_used="error.html"
    )
    assert re.search("Resource not found", resp.content.decode("utf-8"))


def test_browse_content_invalid_origin(client):
    url = reverse(
        "browse-content",
        query_params={
            "origin_url": "http://invalid-origin",
            "path": "/invalid/file/path",
        },
    )

    resp = check_html_get_response(
        client, url, status_code=404, template_used="error.html"
    )
    assert re.search("Resource not found", resp.content.decode("utf-8"))


def test_origin_content_view(
    client, archive_data, swh_scheduler, origin_with_multiple_visits
):
    origin_visits = archive_data.origin_visit_get(origin_with_multiple_visits["url"])

    def _get_archive_data(visit_idx):
        snapshot = archive_data.snapshot_get(origin_visits[visit_idx]["snapshot"])
        head_rev_id = archive_data.snapshot_get_head(snapshot)
        head_rev = archive_data.revision_get(head_rev_id)
        dir_content = archive_data.directory_ls(head_rev["directory"])
        dir_files = [e for e in dir_content if e["type"] == "file"]
        dir_file = random.choice(dir_files)
        branches, releases, _ = process_snapshot_branches(snapshot)
        return {
            "branches": branches,
            "releases": releases,
            "root_dir_sha1": head_rev["directory"],
            "content": get_content(dir_file["checksums"]["sha1"]),
            "visit": origin_visits[visit_idx],
            "snapshot_sizes": archive_data.snapshot_count_branches(snapshot["id"]),
        }

    tdata = _get_archive_data(-1)

    _origin_content_view_test_helper(
        client,
        archive_data,
        origin_with_multiple_visits,
        origin_visits[-1],
        tdata["snapshot_sizes"],
        tdata["branches"],
        tdata["releases"],
        tdata["root_dir_sha1"],
        tdata["content"],
    )

    _origin_content_view_test_helper(
        client,
        archive_data,
        origin_with_multiple_visits,
        origin_visits[-1],
        tdata["snapshot_sizes"],
        tdata["branches"],
        tdata["releases"],
        tdata["root_dir_sha1"],
        tdata["content"],
        timestamp=tdata["visit"]["date"],
    )

    _origin_content_view_test_helper(
        client,
        archive_data,
        origin_with_multiple_visits,
        origin_visits[-1],
        tdata["snapshot_sizes"],
        tdata["branches"],
        tdata["releases"],
        tdata["root_dir_sha1"],
        tdata["content"],
        snapshot_id=tdata["visit"]["snapshot"],
    )

    tdata = _get_archive_data(0)

    _origin_content_view_test_helper(
        client,
        archive_data,
        origin_with_multiple_visits,
        origin_visits[0],
        tdata["snapshot_sizes"],
        tdata["branches"],
        tdata["releases"],
        tdata["root_dir_sha1"],
        tdata["content"],
        visit_id=tdata["visit"]["visit"],
    )

    _origin_content_view_test_helper(
        client,
        archive_data,
        origin_with_multiple_visits,
        origin_visits[0],
        tdata["snapshot_sizes"],
        tdata["branches"],
        tdata["releases"],
        tdata["root_dir_sha1"],
        tdata["content"],
        snapshot_id=tdata["visit"]["snapshot"],
    )


def _origin_content_view_test_helper(
    client,
    archive_data,
    origin_info,
    origin_visit,
    snapshot_sizes,
    origin_branches,
    origin_releases,
    root_dir_sha1,
    content,
    visit_id=None,
    timestamp=None,
    snapshot_id=None,
):
    content_path = "/".join(content["path"].split("/")[1:])

    if not visit_id and not snapshot_id:
        visit_id = origin_visit["visit"]

    query_params = {"origin_url": origin_info["url"], "path": content_path}

    if timestamp:
        query_params["timestamp"] = timestamp

    if visit_id:
        query_params["visit_id"] = visit_id
    elif snapshot_id:
        query_params["snapshot"] = snapshot_id

    url = reverse(
        "browse-content",
        url_args={"query_string": f"sha1_git:{content['sha1_git']}"},
        query_params=query_params,
    )

    resp = check_html_get_response(
        client, url, status_code=200, template_used="browse-content.html"
    )

    assert type(content["data"]) is str

    assert_contains(resp, '<code class="%s">' % content["hljs_language"])
    assert_contains(resp, escape(content["data"]))

    split_path = content_path.split("/")

    filename = split_path[-1]
    path = content_path.replace(filename, "")[:-1]

    path_info = gen_path_info(path)

    del query_params["path"]

    if timestamp:
        query_params["timestamp"] = format_utc_iso_date(
            parse_iso8601_date_to_utc(timestamp).isoformat(), "%Y-%m-%dT%H:%M:%SZ"
        )

    root_dir_url = reverse(
        "browse-directory",
        url_args={"sha1_git": root_dir_sha1},
        query_params=query_params,
    )

    assert_contains(resp, '<li class="swh-path">', count=len(path_info) + 1)

    assert_contains(resp, '<a href="%s">%s</a>' % (root_dir_url, root_dir_sha1[:7]))

    for p in path_info:
        query_params["path"] = p["path"]
        dir_url = reverse("browse-origin-directory", query_params=query_params)
        assert_contains(resp, '<a href="%s">%s</a>' % (dir_url, p["name"]))

    assert_contains(resp, "<li>%s</li>" % filename)

    query_string = "sha1_git:" + content["sha1_git"]

    url_raw = reverse(
        "browse-content-raw",
        url_args={"query_string": query_string},
        query_params={"filename": filename},
    )
    assert_contains(resp, url_raw)

    if "path" in query_params:
        del query_params["path"]

    origin_branches_url = reverse("browse-origin-branches", query_params=query_params)

    assert_contains(resp, f'href="{escape(origin_branches_url)}"')
    assert_contains(resp, f"Branches ({snapshot_sizes['revision']})")

    origin_releases_url = reverse("browse-origin-releases", query_params=query_params)

    assert_contains(resp, f'href="{escape(origin_releases_url)}">')
    assert_contains(resp, f"Releases ({snapshot_sizes['release']})")

    assert_contains(resp, '<li class="swh-branch">', count=len(origin_branches))

    query_params["path"] = content_path

    for branch in origin_branches:
        root_dir_branch_url = reverse(
            "browse-origin-content",
            query_params={"branch": branch["name"], **query_params},
        )

        assert_contains(resp, '<a href="%s">' % root_dir_branch_url)

    assert_contains(resp, '<li class="swh-release"', count=len(origin_releases))

    query_params["branch"] = None
    for release in origin_releases:
        root_dir_release_url = reverse(
            "browse-origin-content",
            query_params={"release": release["name"], **query_params},
        )

        assert_contains(resp, '<a href="%s">' % root_dir_release_url)

    url = reverse(
        "browse-content",
        url_args={"query_string": query_string},
        query_params=query_params,
    )

    resp = check_html_get_response(
        client, url, status_code=200, template_used="browse-content.html"
    )

    snapshot = archive_data.snapshot_get(origin_visit["snapshot"])
    head_rev_id = archive_data.snapshot_get_head(snapshot)

    swhid_context = {
        "origin": origin_info["url"],
        "visit": gen_swhid(ObjectType.SNAPSHOT, snapshot["id"]),
        "anchor": gen_swhid(ObjectType.REVISION, head_rev_id),
        "path": f"/{content_path}",
    }

    swh_cnt_id = gen_swhid(
        ObjectType.CONTENT, content["sha1_git"], metadata=swhid_context
    )
    swh_cnt_id_url = reverse("browse-swhid", url_args={"swhid": swh_cnt_id})
    assert_contains(resp, swh_cnt_id)
    assert_contains(resp, swh_cnt_id_url)

    assert_contains(resp, "swh-take-new-snapshot")

    _check_origin_link(resp, origin_info["url"])

    assert_not_contains(resp, "swh-metadata-popover")

    content_download_url = reverse(
        "api-1-content-raw",
        url_args={"q": f"sha1_git:{content['sha1_git']}"},
        query_params={"filename": filename},
    )
    assert_contains(resp, content_download_url)


def _check_origin_link(resp, origin_url):
    browse_origin_url = reverse(
        "browse-origin", query_params={"origin_url": origin_url}
    )
    assert_contains(resp, f'href="{browse_origin_url}"')


@pytest.mark.django_db
@pytest.mark.parametrize("staff_user_logged_in", [False, True])
def test_browse_content_snapshot_context_release_directory_target(
    client, staff_user, archive_data, directory_with_files, staff_user_logged_in
):
    if staff_user_logged_in:
        client.force_login(staff_user)

    release_name = "v1.0.0"
    release = Release(
        name=release_name.encode(),
        message=f"release {release_name}".encode(),
        target=hash_to_bytes(directory_with_files),
        target_type=ReleaseTargetType.DIRECTORY,
        synthetic=True,
    )
    archive_data.release_add([release])

    snapshot = Snapshot(
        branches={
            release_name.encode(): SnapshotBranch(
                target=release.id, target_type=SnapshotTargetType.RELEASE
            ),
        },
    )
    archive_data.snapshot_add([snapshot])

    dir_content = archive_data.directory_ls(directory_with_files)
    file_entry = random.choice(
        [entry for entry in dir_content if entry["type"] == "file"]
    )

    sha1_git = file_entry["checksums"]["sha1_git"]

    browse_url = reverse(
        "browse-content",
        url_args={"query_string": f"sha1_git:{sha1_git}"},
        query_params={
            "path": file_entry["name"],
            "release": release_name,
            "snapshot": snapshot.id.hex(),
        },
    )

    check_html_get_response(
        client, browse_url, status_code=200, template_used="browse-content.html"
    )


@given(new_origin())
def test_browse_content_from_dir_with_origin_context_and_breadcrumbs(
    archive_data, client, content_text, content_image_type, new_origin
):
    # create a new origin with a snapshot including two releases targeting directories
    # containing a single file and a HEAD branch targeting the first release

    first_release_subdir = Directory(
        entries=(
            DirectoryEntry(
                name=b"text_file",
                type="file",
                target=hash_to_bytes(content_text["sha1_git"]),
                perms=DentryPerms.content,
            ),
        )
    )

    first_subdir_name = ("1" * 40).encode()
    second_subdir_name = b"foo-" + first_subdir_name

    first_release_dir = Directory(
        entries=(
            DirectoryEntry(
                name=first_subdir_name,
                type="dir",
                target=first_release_subdir.id,
                perms=DentryPerms.directory,
            ),
            DirectoryEntry(
                name=second_subdir_name,
                type="dir",
                target=first_release_subdir.id,
                perms=DentryPerms.directory,
            ),
        )
    )

    second_release_dir = Directory(
        entries=(
            DirectoryEntry(
                name=b"image_file",
                type="file",
                target=hash_to_bytes(content_image_type["sha1_git"]),
                perms=DentryPerms.content,
            ),
        )
    )
    archive_data.directory_add([first_release_dir, second_release_dir])

    first_release_name = "v1.0.0"
    first_release = Release(
        name=first_release_name.encode(),
        message=f"release {first_release_name}".encode(),
        target=first_release_dir.id,
        target_type=ReleaseTargetType.DIRECTORY,
        synthetic=True,
    )
    second_release_name = "v2.0.0"
    second_release = Release(
        name=second_release_name.encode(),
        message=f"release {second_release_name}".encode(),
        target=first_release_dir.id,
        target_type=ReleaseTargetType.DIRECTORY,
        synthetic=True,
    )
    archive_data.release_add([first_release, second_release])

    snapshot = Snapshot(
        branches={
            first_release_name.encode(): SnapshotBranch(
                target=first_release.id, target_type=SnapshotTargetType.RELEASE
            ),
            second_release_name.encode(): SnapshotBranch(
                target=second_release.id, target_type=SnapshotTargetType.RELEASE
            ),
            b"HEAD": SnapshotBranch(
                target=first_release_name.encode(), target_type=SnapshotTargetType.ALIAS
            ),
        },
    )
    archive_data.snapshot_add([snapshot])

    archive_data.origin_add([new_origin])
    visit_date = now()
    visit = archive_data.origin_visit_add(
        [
            OriginVisit(
                origin=new_origin.url,
                date=visit_date,
                type="git",
            )
        ]
    )[0]
    visit_status = OriginVisitStatus(
        origin=new_origin.url,
        visit=visit.visit,
        date=visit_date + timedelta(minutes=5),
        status="full",
        snapshot=snapshot.id,
    )
    archive_data.origin_visit_status_add([visit_status])

    # browse content in the second release with origin context,
    # the id of the directory containing the file is prepended to the
    # path query parameter in order for breadcrumbs to be displayed
    url = reverse(
        "browse-content",
        url_args={"query_string": f"sha1_git:{content_image_type['sha1_git']}"},
        query_params={
            "path": f"{second_release_dir.id.hex()}/image_file",
            "origin_url": new_origin.url,
        },
    )

    check_html_get_response(
        client, url, status_code=200, template_used="browse-content.html"
    )

    # check paths containing a sha1 sum can still be browsed
    for subdir_name in (first_subdir_name, second_subdir_name):
        url = reverse(
            "browse-content",
            url_args={"query_string": f"sha1_git:{content_text['sha1_git']}"},
            query_params={
                "path": f"{subdir_name.decode()}/text_file",
                "origin_url": new_origin.url,
            },
        )

        check_html_get_response(
            client, url, status_code=200, template_used="browse-content.html"
        )


@override_settings(RATELIMIT_ENABLE=True)
@pytest.mark.parametrize("view_name", ["browse-content", "browse-content-raw"])
def test_browse_content_rate_limit(client, content_text, view_name):
    url = reverse(
        view_name,
        url_args={"query_string": f"sha1_git:{content_text['sha1_git']}"},
    )

    check_http_get_response(client, url, status_code=200)
    check_http_get_response(client, url, status_code=429)


@override_settings(RATELIMIT_ENABLE=True)
@pytest.mark.parametrize(
    "forwarded_for",
    ["1234:1234:1234::123", "123.123.123.123", "1234:1234:1234::123, 123.123.123.123"],
)
def test_browse_content_rate_limit_forwarded_for(client, content_text, forwarded_for):
    url = reverse(
        "browse-content",
        url_args={"query_string": f"sha1_git:{content_text['sha1_git']}"},
    )

    check_http_get_response(
        client, url, status_code=200, HTTP_X_ORIGINAL_FORWARDED_FOR=forwarded_for
    )
    check_http_get_response(
        client, url, status_code=429, HTTP_X_ORIGINAL_FORWARDED_FOR=forwarded_for
    )


def test_browse_content_failed_encoding_detection(
    client, content_text_non_utf8, mocker
):
    # simulate charset_normalizer.detect failure
    detect = mocker.patch("charset_normalizer.detect")
    detect.return_value = {"confidence": None, "encoding": None, "language": ""}

    url = reverse(
        "browse-content",
        url_args={"query_string": f"sha1_git:{content_text_non_utf8['sha1_git']}"},
    )

    # content should be rendered even if encoding detection failed
    check_http_get_response(client, url, status_code=200)
    detect.assert_called()


def test_browse_content_without_root_dir_info_in_path(
    archive_data,
    client,
    content_text,
):
    new_origin = Origin(url="https://git.example.org/test/browse/no/root/dir")
    root_dir = Directory(
        entries=(
            DirectoryEntry(
                name=b"text_file",
                type="file",
                target=hash_to_bytes(content_text["sha1_git"]),
                perms=DentryPerms.content,
            ),
        )
    )
    archive_data.directory_add([root_dir])

    release_name = "v1.0.0"
    release = Release(
        name=release_name.encode(),
        message=f"release {release_name}".encode(),
        target=root_dir.id,
        target_type=ReleaseTargetType.DIRECTORY,
        synthetic=True,
    )
    archive_data.release_add([release])

    snapshot = Snapshot(
        branches={
            release_name.encode(): SnapshotBranch(
                target=release.id, target_type=SnapshotTargetType.RELEASE
            ),
            b"HEAD": SnapshotBranch(
                target=release_name.encode(), target_type=SnapshotTargetType.ALIAS
            ),
        },
    )

    archive_data.snapshot_add([snapshot])
    archive_data.origin_add([new_origin])

    visit_date = now()
    visit = archive_data.origin_visit_add(
        [
            OriginVisit(
                origin=new_origin.url,
                date=visit_date,
                type="git",
            )
        ]
    )[0]
    visit_status = OriginVisitStatus(
        origin=new_origin.url,
        visit=visit.visit,
        date=visit_date + timedelta(minutes=5),
        status="full",
        snapshot=snapshot.id,
    )
    archive_data.origin_visit_status_add([visit_status])

    url = reverse(
        "browse-content",
        url_args={"query_string": f"sha1_git:{content_text['sha1_git']}"},
        query_params={
            "path": "/text_file",
            "origin_url": new_origin.url,
            "snapshot": snapshot.id.hex(),
        },
    )

    check_html_get_response(
        client, url, status_code=200, template_used="browse-content.html"
    )


def test_browse_contents_diff(client, archive_data):
    content_from = Content.from_data(
        b"""\
import os

print("number of CPUs: ", os.cpu_count())
"""
    )

    content_to = Content.from_data(
        b"""\
import os

print("number of CPUs: ", os.cpu_count())
print("load average: ", os.getloadavg())
"""
    )

    archive_data.content_add([content_from, content_to])

    url = reverse(
        "diff-contents",
        url_args={
            "from_query_string": f"sha1_git:{content_from.sha1_git.hex()}",
            "to_query_string": f"sha1_git:{content_to.sha1_git.hex()}",
        },
    )

    resp = check_http_get_response(client, url, status_code=200)

    assert (
        resp.json()["diff_str"]
        == """\
@@ -1,3 +1,4 @@
 import os
 
 print("number of CPUs: ", os.cpu_count())
+print("load average: ", os.getloadavg())
"""  # noqa
    )


def test_browse_contents_diff_remove_form_feeds(client, archive_data):
    content_from = Content.from_data(
        b"""\
import os
\f
print("number of CPUs: ", os.cpu_count())
"""
    )

    content_to = Content.from_data(
        b"""\
import os
\f
print("number of CPUs: ", os.cpu_count())
print("load average: ", os.getloadavg())
"""
    )

    archive_data.content_add([content_from, content_to])

    url = reverse(
        "diff-contents",
        url_args={
            "from_query_string": f"sha1_git:{content_from.sha1_git.hex()}",
            "to_query_string": f"sha1_git:{content_to.sha1_git.hex()}",
        },
    )

    resp = check_http_get_response(client, url, status_code=200)

    assert (
        resp.json()["diff_str"]
        == """\
@@ -1,3 +1,4 @@
 import os
 
 print("number of CPUs: ", os.cpu_count())
+print("load average: ", os.getloadavg())
"""  # noqa
    )


def test_browse_json_contents_diff(client, archive_data):
    content_from = Content.from_data(
        b"""\
{
    "key": "foo"
}
"""
    )

    content_to = Content.from_data(
        b"""\
{
    "key": "foo",
    "value": "bar"
}
"""
    )

    archive_data.content_add([content_from, content_to])

    url = reverse(
        "diff-contents",
        url_args={
            "from_query_string": f"sha1_git:{content_from.sha1_git.hex()}",
            "to_query_string": f"sha1_git:{content_to.sha1_git.hex()}",
        },
    )

    resp = check_http_get_response(client, url, status_code=200)

    assert (
        resp.json()["diff_str"]
        == """\
@@ -1,3 +1,4 @@
 {
-    "key": "foo"
+    "key": "foo",
+    "value": "bar"
 }
"""
    )


def test_browse_contents_diff_for_added_content(client, archive_data):
    new_content = Content.from_data(b"foo\n")

    archive_data.content_add([new_content])

    url = reverse(
        "diff-contents",
        url_args={
            "from_query_string": "",
            "to_query_string": f"sha1_git:{new_content.sha1_git.hex()}",
        },
    )

    resp = check_http_get_response(client, url, status_code=200)

    assert (
        resp.json()["diff_str"]
        == """\
@@ -0,0 +1 @@
+foo
"""
    )



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


APP_NAME = "swh_web_mailmap"

MIGRATION_0001 = "0001_initial"


def test_mailmap_django_app(migrator):
    state = migrator.apply_tested_migration((APP_NAME, MIGRATION_0001))
    UserMailmap = state.apps.get_model(APP_NAME, "UserMailmap")
    assert UserMailmap



# Copyright (C) 2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import hashlib
from urllib.parse import urlparse, urlunparse

from django_ratelimit.decorators import ratelimit

from django.http import HttpRequest, HttpResponse, HttpResponseRedirect, QueryDict
from django.shortcuts import render

from swh.web.utils import (
    archive,
    cache_get,
    cache_set,
    origin_visit_types,
    reverse,
    strtobool,
)
from swh.web.utils.identifiers import resolve_swhid


@ratelimit(key="user_or_ip", rate="60/m")
def browse_search_view(request: HttpRequest) -> HttpResponse:
    origins: list = []
    total_results: int | None = -1
    previous_page_url = None
    next_page_url = None
    if "no_js" in request.GET:
        # apply same client side processing when javascript is disabled

        # get search parameter values
        query = request.GET.get("q", "").strip()
        with_visit = strtobool(request.GET.get("with_visit", "on"))
        with_content = strtobool(request.GET.get("with_content", "on"))
        visit_type = request.GET.get("visit_type", "any")
        use_ql = strtobool(request.GET.get("use_ql", "off"))
        search_in_metadata = strtobool(request.GET.get("search_in_metadata", "off"))
        page_token = request.GET.get("page_token")
        prev_page_token = None
        next_page_token = None
        limit = 100

        # if a SWHID is provided, try to resolve it and redirect to object page
        if query.startswith("swh:"):
            swhid_resolved = resolve_swhid(query)
            if swhid_resolved["browse_url"]:
                return HttpResponseRedirect(swhid_resolved["browse_url"])

        # if an origin URL is provided and is archived, redirect to its browsing page
        try:
            if origin_info := archive.lookup_origin_visit_latest(
                query,
                require_snapshot=True,
                type=visit_type if visit_type != "any" else None,
            ):
                return HttpResponseRedirect(
                    reverse(
                        "browse-origin",
                        query_params={"origin_url": origin_info["origin"]},
                    )
                )
        except Exception:
            pass

        if search_in_metadata:
            origins = list(
                archive.search_origin_metadata(
                    query,
                    limit,
                    return_metadata=False,
                )
            )
        else:
            origins, next_page_token, total_results = archive.search_origin(
                url_pattern=query,
                use_ql=use_ql,
                limit=limit,
                with_visit=with_visit,
                with_content=with_content,
                visit_types=[visit_type] if visit_type not in (None, "any") else None,
                page_token=page_token,
            )
            total_results = total_results or 0

        # cache page tokens for previous and next links
        key = "origin_search" + "_".join(
            map(str, [query, with_visit, with_content, visit_type, use_ql, limit])
        )
        cache_key = hashlib.md5(key.encode()).hexdigest()
        page_tokens = cache_get(cache_key) or []
        if next_page_token and (not page_tokens or page_tokens[-1] == page_token):
            page_tokens.append(next_page_token)
        cache_set(cache_key, page_tokens)

        # compute previous and next links if any
        current_url = request.build_absolute_uri()
        parsed_url = urlparse(current_url)
        query_dict = QueryDict(parsed_url.query, mutable=True)

        if page_token:
            pos = page_tokens.index(page_token)
            if pos > 0:
                prev_page_token = page_tokens[pos - 1]
                query_dict["page_token"] = prev_page_token
                previous_page_url = urlunparse(
                    parsed_url._replace(query=query_dict.urlencode(safe="/;:"))
                )
            elif pos == 0:
                query_dict.pop("page_token", None)
                previous_page_url = urlunparse(
                    parsed_url._replace(query=query_dict.urlencode(safe="/;:"))
                )

        if next_page_token:
            query_dict["page_token"] = next_page_token
            next_page_url = urlunparse(
                parsed_url._replace(query=query_dict.urlencode(safe="/;:"))
            )

    return render(
        request,
        "browse-search.html",
        {
            "heading": "Search software origins to browse",
            "visit_types": origin_visit_types(use_cache=True),
            "total_results": total_results,
            "origins": origins,
            "previous_page_url": previous_page_url,
            "next_page_url": next_page_url,
        },
    )
